import re
import json
import logging
from unittest.mock import AsyncMock, patch
from pytest import mark, fixture, raises
from aiohttp import WSMessage, WSMsgType

import proximl.cloudbender.nodes as specimen
from proximl.exceptions import (
    ApiError,
    SpecificationError,
    ProxiMLException,
)

pytestmark = [mark.sdk, mark.unit, mark.cloudbender, mark.nodes]


@fixture
def nodes(mock_proximl):
    yield specimen.Nodes(mock_proximl)


@fixture
def node(mock_proximl):
    yield specimen.Node(
        mock_proximl,
        provider_uuid="1",
        region_uuid="a",
        rig_uuid="x",
        type="permanent",
        service="compute",
        friendly_name="hq-a100-01",
        hostname="hq-a100-01",
        status="active",
        online=True,
        maintenance_mode=False,
    )


class RegionsTests:
    @mark.asyncio
    async def test_get_node(
        self,
        nodes,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await nodes.get("1234", "5687", "91011")
        mock_proximl._query.assert_called_once_with(
            "/provider/1234/region/5687/node/91011", "GET", {}
        )

    @mark.asyncio
    async def test_list_nodes(
        self,
        nodes,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await nodes.list("1234", "5687")
        mock_proximl._query.assert_called_once_with(
            "/provider/1234/region/5687/node", "GET", {}
        )

    @mark.asyncio
    async def test_remove_node(
        self,
        nodes,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await nodes.remove("1234", "4567", "8910")
        mock_proximl._query.assert_called_once_with(
            "/provider/1234/region/4567/node/8910", "DELETE", {}
        )

    @mark.asyncio
    async def test_create_node(self, nodes, mock_proximl):
        requested_config = dict(
            provider_uuid="provider-id-1",
            region_uuid="region-id-1",
            friendly_name="phys-node",
            hostname="phys-node",
            minion_id="asdf",
        )
        expected_payload = dict(
            friendly_name="phys-node",
            hostname="phys-node",
            minion_id="asdf",
            type="permanent",
            service="compute",
        )
        api_response = {
            "provider_uuid": "provider-id-1",
            "region_uuid": "region-id-1",
            "rig_uuid": "rig-id-1",
            "name": "phys-node",
            "type": "permanent",
            "service": "compute",
            "status": "new",
            "online": False,
            "maintenance_mode": True,
            "createdAt": "2020-12-31T23:59:59.000Z",
        }

        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await nodes.create(**requested_config)
        mock_proximl._query.assert_called_once_with(
            "/provider/provider-id-1/region/region-id-1/node",
            "POST",
            None,
            expected_payload,
        )
        assert response.id == "rig-id-1"


class nodeTests:
    def test_node_properties(self, node):
        assert isinstance(node.id, str)
        assert isinstance(node.provider_uuid, str)
        assert isinstance(node.region_uuid, str)
        assert isinstance(node.type, str)
        assert isinstance(node.service, str)
        assert isinstance(node.name, str)
        assert isinstance(node.hostname, str)
        assert isinstance(node.status, str)
        assert isinstance(node.online, bool)
        assert isinstance(node.maintenance_mode, bool)

    def test_node_str(self, node):
        string = str(node)
        regex = r"^{.*\"rig_uuid\": \"" + node.id + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_node_repr(self, node):
        string = repr(node)
        regex = (
            r"^Node\( proximl , \*\*{.*'rig_uuid': '" + node.id + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_node_bool(self, node, mock_proximl):
        empty_node = specimen.Node(mock_proximl)
        assert bool(node)
        assert not bool(empty_node)

    @mark.asyncio
    async def test_node_remove(self, node, mock_proximl):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await node.remove()
        mock_proximl._query.assert_called_once_with(
            "/provider/1/region/a/node/x", "DELETE"
        )

    @mark.asyncio
    async def test_node_refresh(self, node, mock_proximl):
        api_response = {
            "provider_uuid": "provider-id-1",
            "region_uuid": "region-id-1",
            "rig_uuid": "rig-id-1",
            "name": "phys-node",
            "type": "permanent",
            "service": "compute",
            "status": "new",
            "online": False,
            "maintenance_mode": True,
            "createdAt": "2020-12-31T23:59:59.000Z",
        }
        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await node.refresh()
        mock_proximl._query.assert_called_once_with(
            f"/provider/1/region/a/node/x", "GET"
        )
        assert node.id == "rig-id-1"
        assert response.id == "rig-id-1"

    @mark.asyncio
    async def test_node_toggle_maintenance(self, node, mock_proximl):
        api_response = None
        mock_proximl._query = AsyncMock(return_value=api_response)
        await node.toggle_maintenance()
        mock_proximl._query.assert_called_once_with(
            "/provider/1/region/a/node/x/maintenance", "PATCH"
        )

    @mark.asyncio
    async def test_node_run_action(self, node, mock_proximl):
        api_response = None
        mock_proximl._query = AsyncMock(return_value=api_response)
        await node.run_action(command="report")
        mock_proximl._query.assert_called_once_with(
            "/provider/1/region/a/node/x/action",
            "POST",
            None,
            dict(command="report"),
        )



import re
import json
import logging
from unittest.mock import AsyncMock, patch
from pytest import mark, fixture, raises
from aiohttp import WSMessage, WSMsgType

import proximl.cloudbender.datastores as specimen
from proximl.exceptions import (
    ApiError,
    SpecificationError,
    ProxiMLException,
)

pytestmark = [mark.sdk, mark.unit, mark.cloudbender, mark.datastores]


@fixture
def datastores(mock_proximl):
    yield specimen.Datastores(mock_proximl)


@fixture
def datastore(mock_proximl):
    yield specimen.Datastore(
        mock_proximl,
        provider_uuid="1",
        region_uuid="a",
        store_id="x",
        name="On-Prem Datastore",
        type="nfs",
        uri="192.168.0.50",
        root="/exports",
    )


class RegionsTests:
    @mark.asyncio
    async def test_get_datastore(
        self,
        datastores,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await datastores.get("1234", "5687", "91011")
        mock_proximl._query.assert_called_once_with(
            "/provider/1234/region/5687/datastore/91011", "GET", {}
        )

    @mark.asyncio
    async def test_list_datastores(
        self,
        datastores,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await datastores.list("1234", "5687")
        mock_proximl._query.assert_called_once_with(
            "/provider/1234/region/5687/datastore", "GET", {}
        )

    @mark.asyncio
    async def test_remove_datastore(
        self,
        datastores,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await datastores.remove("1234", "4567", "8910")
        mock_proximl._query.assert_called_once_with(
            "/provider/1234/region/4567/datastore/8910", "DELETE", {}
        )

    @mark.asyncio
    async def test_create_datastore(self, datastores, mock_proximl):
        requested_config = dict(
            provider_uuid="provider-id-1",
            region_uuid="region-id-1",
            name="On-Prem Datastore",
            type="nfs",
            uri="192.168.0.50",
            root="/exports",
        )
        expected_payload = dict(
            name="On-Prem Datastore",
            type="nfs",
            uri="192.168.0.50",
            root="/exports",
        )
        api_response = {
            "provider_uuid": "provider-id-1",
            "region_uuid": "region-id-1",
            "store_id": "store-id-1",
            "name": "On-Prem Datastore",
            "type": "nfs",
            "uri": "192.168.0.50",
            "root": "/exports",
            "createdAt": "2020-12-31T23:59:59.000Z",
        }

        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await datastores.create(**requested_config)
        mock_proximl._query.assert_called_once_with(
            "/provider/provider-id-1/region/region-id-1/datastore",
            "POST",
            None,
            expected_payload,
        )
        assert response.id == "store-id-1"


class datastoreTests:
    def test_datastore_properties(self, datastore):
        assert isinstance(datastore.id, str)
        assert isinstance(datastore.provider_uuid, str)
        assert isinstance(datastore.region_uuid, str)
        assert isinstance(datastore.type, str)
        assert isinstance(datastore.name, str)

    def test_datastore_str(self, datastore):
        string = str(datastore)
        regex = r"^{.*\"store_id\": \"" + datastore.id + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_datastore_repr(self, datastore):
        string = repr(datastore)
        regex = (
            r"^Datastore\( proximl , \*\*{.*'store_id': '" + datastore.id + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_datastore_bool(self, datastore, mock_proximl):
        empty_datastore = specimen.Datastore(mock_proximl)
        assert bool(datastore)
        assert not bool(empty_datastore)

    @mark.asyncio
    async def test_datastore_remove(self, datastore, mock_proximl):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await datastore.remove()
        mock_proximl._query.assert_called_once_with(
            "/provider/1/region/a/datastore/x", "DELETE"
        )

    @mark.asyncio
    async def test_datastore_refresh(self, datastore, mock_proximl):
        api_response = {
            "provider_uuid": "provider-id-1",
            "region_uuid": "region-id-1",
            "store_id": "store-id-1",
            "name": "On-Prem Datastore",
            "type": "nfs",
            "uri": "192.168.0.50",
            "root": "/exports",
            "createdAt": "2020-12-31T23:59:59.000Z",
        }
        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await datastore.refresh()
        mock_proximl._query.assert_called_once_with(
            f"/provider/1/region/a/datastore/x", "GET"
        )
        assert datastore.id == "store-id-1"
        assert response.id == "store-id-1"



import json
import logging
import asyncio
import math
from datetime import datetime

from proximl.exceptions import (
    ApiError,
    SpecificationError,
    ProxiMLException,
    ProviderError,
)


class Providers(object):
    def __init__(self, proximl):
        self.proximl = proximl

    async def get(self, id):
        resp = await self.proximl._query(f"/provider/{id}", "GET")
        return Provider(self.proximl, **resp)

    async def list(self):
        resp = await self.proximl._query(f"/provider", "GET")
        providers = [Provider(self.proximl, **provider) for provider in resp]
        return providers

    async def enable(self, type, **kwargs):
        data = dict(type=type, **kwargs)
        payload = {k: v for k, v in data.items() if v is not None}
        logging.info(f"Enabling Provider {type}")
        resp = await self.proximl._query("/provider", "POST", None, payload)
        provider = Provider(self.proximl, **resp)
        logging.info(f"Enabled Provider {type} with id {provider.id}")

        return provider

    async def remove(self, id):
        await self.proximl._query(f"/provider/{id}", "DELETE")


class Provider:
    def __init__(self, proximl, **kwargs):
        self.proximl = proximl
        self._provider = kwargs
        self._id = self._provider.get("provider_uuid")
        self._type = self._provider.get("type")
        self._status = self._provider.get("status")
        self._credits = self._provider.get("credits")

    @property
    def id(self) -> str:
        return self._id

    @property
    def type(self) -> str:
        return self._type

    @property
    def status(self) -> str:
        return self._status

    @property
    def credits(self) -> float:
        return self._credits

    def __str__(self):
        return json.dumps({k: v for k, v in self._provider.items()})

    def __repr__(self):
        return f"Provider( proximl , **{self._provider.__repr__()})"

    def __bool__(self):
        return bool(self._id)

    async def remove(self):
        await self.proximl._query(f"/provider/{self._id}", "DELETE")

    async def refresh(self):
        resp = await self.proximl._query(
            f"/provider/{self._id}",
            "GET",
        )
        self.__init__(self.proximl, **resp)
        return self

    async def wait_for(self, status, timeout=300):
        if self.status == status:
            return
        valid_statuses = ["ready", "archived"]
        if not status in valid_statuses:
            raise SpecificationError(
                "status",
                f"Invalid wait_for status {status}.  Valid statuses are: {valid_statuses}",
            )
        MAX_TIMEOUT = 24 * 60 * 60
        if timeout > MAX_TIMEOUT:
            raise SpecificationError(
                "timeout",
                f"timeout must be less than {MAX_TIMEOUT} seconds.",
            )

        POLL_INTERVAL_MIN = 5
        POLL_INTERVAL_MAX = 60
        POLL_INTERVAL = max(min(timeout / 60, POLL_INTERVAL_MAX), POLL_INTERVAL_MIN)
        retry_count = math.ceil(timeout / POLL_INTERVAL)
        count = 0
        while count < retry_count:
            await asyncio.sleep(POLL_INTERVAL)
            try:
                await self.refresh()
            except ApiError as e:
                if status == "archived" and e.status == 404:
                    return
                raise e
            if self.status in ["errored", "failed"]:
                raise ProviderError(self.status, self)
            if self.status == status:
                return self
            else:
                count += 1
                logging.debug(f"self: {self}, retry count {count}")

        raise ProxiMLException(f"Timeout waiting for {status}")



import re
import json
import logging
from unittest.mock import AsyncMock, patch
from pytest import mark, fixture, raises
from aiohttp import WSMessage, WSMsgType

import proximl.cloudbender.providers as specimen
from proximl.exceptions import (
    ApiError,
    SpecificationError,
    ProxiMLException,
)

pytestmark = [mark.sdk, mark.unit, mark.cloudbender, mark.providers]


@fixture
def providers(mock_proximl):
    yield specimen.Providers(mock_proximl)


@fixture
def provider(mock_proximl):
    yield specimen.Provider(
        mock_proximl,
        customer_uuid="a",
        provider_uuid="1",
        type="physical",
        payment_mode="credits",
        createdAt="2020-12-31T23:59:59.000Z",
        credits=0.0,
    )


class ProvidersTests:
    @mark.asyncio
    async def test_get_provider(
        self,
        providers,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await providers.get("1234")
        mock_proximl._query.assert_called_once_with("/provider/1234", "GET")

    @mark.asyncio
    async def test_list_providers(
        self,
        providers,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await providers.list()
        mock_proximl._query.assert_called_once_with("/provider", "GET")

    @mark.asyncio
    async def test_remove_provider(
        self,
        providers,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await providers.remove("4567")
        mock_proximl._query.assert_called_once_with("/provider/4567", "DELETE")

    @mark.asyncio
    async def test_enable_provider_simple(self, providers, mock_proximl):
        requested_config = dict(
            type="physical",
        )
        expected_payload = dict(type="physical")
        api_response = {
            "customer_uuid": "cust-id-1",
            "provider_uuid": "provider-id-1",
            "type": "new provider",
            "credits": 0.0,
            "payment_mode": "credits",
            "createdAt": "2020-12-31T23:59:59.000Z",
        }

        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await providers.enable(**requested_config)
        mock_proximl._query.assert_called_once_with(
            "/provider", "POST", None, expected_payload
        )
        assert response.id == "provider-id-1"


class providerTests:
    def test_provider_properties(self, provider):
        assert isinstance(provider.id, str)
        assert isinstance(provider.type, str)
        assert isinstance(provider.credits, float)

    def test_provider_str(self, provider):
        string = str(provider)
        regex = r"^{.*\"provider_uuid\": \"" + provider.id + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_provider_repr(self, provider):
        string = repr(provider)
        regex = (
            r"^Provider\( proximl , \*\*{.*'provider_uuid': '"
            + provider.id
            + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_provider_bool(self, provider, mock_proximl):
        empty_provider = specimen.Provider(mock_proximl)
        assert bool(provider)
        assert not bool(empty_provider)

    @mark.asyncio
    async def test_provider_remove(self, provider, mock_proximl):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await provider.remove()
        mock_proximl._query.assert_called_once_with("/provider/1", "DELETE")

    @mark.asyncio
    async def test_provider_refresh(self, provider, mock_proximl):
        api_response = {
            "customer_uuid": "cust-id-1",
            "provider_uuid": "provider-id-1",
            "type": "new provider",
            "credits": 0.0,
            "payment_mode": "credits",
            "createdAt": "2020-12-31T23:59:59.000Z",
        }
        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await provider.refresh()
        mock_proximl._query.assert_called_once_with(f"/provider/1", "GET")
        assert provider.id == "provider-id-1"
        assert response.id == "provider-id-1"



import json
import logging
from .datastores import ProjectDatastores
from .data_connectors import ProjectDataConnectors
from .services import ProjectServices
from .credentials import ProjectCredentials
from .secrets import ProjectSecrets
from .members import ProjectMembers


class Projects(object):
    def __init__(self, proximl):
        self.proximl = proximl

    async def get(self, id, **kwargs):
        resp = await self.proximl._query(f"/project/{id}", "GET", kwargs)
        return Project(self.proximl, **resp)

    async def get_current(self, **kwargs):
        resp = await self.proximl._query(
            f"/project/{self.proximl.project}", "GET", kwargs
        )
        return Project(self.proximl, **resp)

    async def list(self, **kwargs):
        resp = await self.proximl._query(f"/project", "GET", kwargs)
        projects = [Project(self.proximl, **project) for project in resp]
        return projects

    async def create(self, name, **kwargs):
        data = dict(
            name=name,
            **kwargs
        )
        payload = {k: v for k, v in data.items() if v is not None}
        logging.info(f"Creating Project {name}")
        resp = await self.proximl._query("/project", "POST", None, payload)
        project = Project(self.proximl, **resp)
        logging.info(f"Created Project {name} with id {project.id}")

        return project

    async def remove(self, id, **kwargs):
        await self.proximl._query(f"/project/{id}", "DELETE", kwargs)


class Project:
    def __init__(self, proximl, **kwargs):
        self.proximl = proximl
        self._entity = kwargs
        self._id = self._entity.get("id")
        self._name = self._entity.get("name")
        self._is_owner = self._entity.get("owner")
        self._owner_name = self._entity.get("owner_name")
        self.datastores = ProjectDatastores(self.proximl, self._id)
        self.data_connectors = ProjectDataConnectors(self.proximl, self._id)
        self.services = ProjectServices(self.proximl, self._id)
        self.credentials = ProjectCredentials(self.proximl, self._id)
        self.secrets = ProjectSecrets(self.proximl, self._id)
        self.members = ProjectMembers(self.proximl, self._id)

    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def is_owner(self) -> bool:
        return self._is_owner

    @property
    def owner_name(self) -> str:
        return self._owner_name

    def __str__(self):
        return json.dumps({k: v for k, v in self._entity.items()})

    def __repr__(self):
        return f"Project( proximl , **{self._entity.__repr__()})"

    def __bool__(self):
        return bool(self._id)

    async def remove(self):
        await self.proximl._query(f"/project/{self._id}", "DELETE")



import json
import logging
from datetime import datetime
from dateutil import parser, tz


class ProjectCredentials(object):
    def __init__(self, proximl, project_id):
        self.proximl = proximl
        self.project_id = project_id

    async def list(self, **kwargs):
        resp = await self.proximl._query(
            f"/project/{self.project_id}/credentials", "GET", kwargs
        )
        credentials = [ProjectCredential(self.proximl, **service) for service in resp]
        return credentials

    async def put(self, type, key_id, secret, tenant=None, **kwargs):
        data = dict(key_id=key_id, secret=secret, tenant=tenant)
        payload = {k: v for k, v in data.items() if v is not None}
        logging.info(f"Creating Project Credential {type}")
        resp = await self.proximl._query(
            f"/project/{self.project_id}/credential/{type}", "PUT", None, payload
        )
        credential = ProjectCredential(self.proximl, **resp)
        logging.info(f"Created Project Credential {type} in project {self.project_id}")

        return credential

    async def remove(self, type, **kwargs):
        await self.proximl._query(
            f"/project/{self.project_id}/credential/{type}", "DELETE", kwargs
        )


class ProjectCredential:
    def __init__(self, proximl, **kwargs):
        self.proximl = proximl
        self._entity = kwargs
        self._type = self._entity.get("type")
        self._project_uuid = self._entity.get("project_uuid")
        self._key_id = self._entity.get("key_id")
        self._updated_at = self._entity.get("updatedAt")

    @property
    def type(self) -> str:
        return self._type

    @property
    def project_uuid(self) -> str:
        return self._project_uuid

    @property
    def key_id(self) -> str:
        return self._key_id

    @property
    def updated_at(self) -> datetime:
        timestamp = parser.isoparse(self._updated_at)
        timezone = tz.tzlocal()
        return timestamp.astimezone(timezone)

    def __str__(self):
        return json.dumps({k: v for k, v in self._entity.items()})

    def __repr__(self):
        return f"ProjectCredential( proximl , **{self._entity.__repr__()})"

    def __bool__(self):
        return bool(self._type)



import click
from proximl.cli import cli, pass_config, search_by_id_name
from proximl.cli.cloudbender import cloudbender


@cloudbender.group()
@pass_config
def provider(config):
    """proxiML CloudBender provider commands."""
    pass


@provider.command()
@pass_config
def list(config):
    """List providers."""
    data = [
        ["ID", "TYPE", "CREDITS"],
        [
            "-" * 80,
            "-" * 80,
            "-" * 80,
        ],
    ]

    providers = config.proximl.run(
        config.proximl.client.cloudbender.providers.list()
    )

    for provider in providers:
        data.append(
            [
                provider.id,
                provider.type,
                f"{provider.credits}",
            ]
        )

    for row in data:
        click.echo(
            "{: >38.36} {: >30.28} {: >15.13}" "".format(*row),
            file=config.stdout,
        )


@provider.command()
@click.argument("type", type=click.STRING)
@pass_config
def enable(config, type):
    """
    Enables a provider.

    Provider is created of the specified type.
    """

    return config.proximl.run(
        config.proximl.client.cloudbender.providers.enable(
            type=type,
        )
    )


@provider.command()
@click.argument("provider", type=click.STRING)
@pass_config
def remove(config, provider):
    """
    Remove a provider.

    PROVIDER may be specified by name or ID, but ID is preferred.
    """
    providers = config.proximl.run(
        config.proximl.client.cloudbender.providers.list()
    )

    found = search_by_id_name(provider, providers)
    if None is found:
        raise click.UsageError("Cannot find specified provider.")

    return config.proximl.run(found.remove())



import asyncio
import click
from proximl.cli import cli, pass_config


@cli.group()
@pass_config
def gpu(config):
    """proxiML GPU commands."""
    pass


@gpu.command()
@pass_config
def list(config):
    """List GPUs."""
    data = [
        ["ID", "NAME", "ABBRV", "CREDITS/HR MIN", "CREDITS/HR MAX"],
        ["-" * 80, "-" * 80, "-" * 80, "-" * 80, "-" * 80],
    ]

    gpus = config.proximl.run(config.proximl.client.gpu_types.list())

    for gpu in gpus:
        data.append(
            [
                gpu.id,
                gpu.name,
                gpu.abbrv,
                f"{gpu.credits_per_hour_min:.2f}",
                f"{gpu.credits_per_hour_max:.2f}",
            ]
        )
    for row in data:
        click.echo(
            "{: >36.34} {: >16.14} {: >10.8} {: >16.14} {: >16.14}"
            "".format(*row),
            file=config.stdout,
        )



import json
from proximl.exceptions import ProxiMLException


class GpuTypes(object):
    def __init__(self, proximl):
        self.proximl = proximl

    async def list(self):
        if not self.proximl.project:
            raise ProxiMLException(
                "Active project not configured.  Run 'proximl configure' to select an active project."
            )
        resp = await self.proximl._query(
            f"/project/{self.proximl.project}/gputypes", "GET"
        )
        gpu_types = [GpuType(self.proximl, **gpu_type) for gpu_type in resp]
        return gpu_types

    async def refresh_gpu_types(self):
        await self.proximl._query(
            f"/project/{self.proximl.project}/gputypes", "PATCH"
        )


class GpuType:
    def __init__(self, proximl, **kwargs):
        self.proximl = proximl
        self._gpu_type = kwargs
        self._id = self._gpu_type.get("id", self._gpu_type.get("gpu_type_id"))
        self._name = self._gpu_type.get("name")
        self._abbrv = self._gpu_type.get("abbrv")
        self._credits_per_hour_min = self._gpu_type.get("price").get("min")
        self._credits_per_hour_max = self._gpu_type.get("price").get("max")

    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def abbrv(self) -> str:
        return self._abbrv

    @property
    def credits_per_hour_min(self) -> float:
        return self._credits_per_hour_min

    @property
    def credits_per_hour_max(self) -> float:
        return self._credits_per_hour_max

    def __str__(self):
        return json.dumps(
            {
                k: v
                for k, v in self._gpu_type.items()
                if k not in ["createdAt", "updatedAt"]
            }
        )

    def __repr__(self):
        return f"GpuType( proximl , **{self._gpu_type.__repr__()})"



import re
import json
import logging
from unittest.mock import AsyncMock, patch
from pytest import mark, fixture, raises
from aiohttp import WSMessage, WSMsgType

import proximl.projects.data_connectors as specimen
from proximl.exceptions import (
    ApiError,
    SpecificationError,
    ProxiMLException,
)

pytestmark = [mark.sdk, mark.unit, mark.projects]


@fixture
def project_data_connectors(mock_proximl):
    yield specimen.ProjectDataConnectors(mock_proximl, project_id="1")


@fixture
def project_data_connector(mock_proximl):
    yield specimen.ProjectDataConnector(
        mock_proximl,
        id="ds-id-1",
        name="connector 1",
        project_uuid="proj-id-1",
        type="custom",
        region_uuid="reg-id-1",
    )


class ProjectDataConnectorsTests:
    @mark.asyncio
    async def test_project_data_connectors_refresh(
        self, project_data_connectors, mock_proximl
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await project_data_connectors.refresh()
        mock_proximl._query.assert_called_once_with(
            "/project/1/data_connectors", "PATCH"
        )

    @mark.asyncio
    async def test_project_data_connectors_list(
        self, project_data_connectors, mock_proximl
    ):
        api_response = [
            {
                "project_uuid": "proj-id-1",
                "region_uuid": "reg-id-1",
                "id": "store-id-1",
                "type": "custom",
                "name": "On-Prem Connection A",
            },
            {
                "project_uuid": "proj-id-1",
                "region_uuid": "reg-id-2",
                "id": "store-id-2",
                "type": "custom",
                "name": "Cloud Connection B",
            },
        ]
        mock_proximl._query = AsyncMock(return_value=api_response)
        resp = await project_data_connectors.list()
        mock_proximl._query.assert_called_once_with(
            "/project/1/data_connectors", "GET", dict()
        )
        assert len(resp) == 2


class ProjectDataConnectorTests:
    def test_project_data_connector_properties(self, project_data_connector):
        assert isinstance(project_data_connector.id, str)
        assert isinstance(project_data_connector.name, str)
        assert isinstance(project_data_connector.project_uuid, str)
        assert isinstance(project_data_connector.type, str)
        assert isinstance(project_data_connector.region_uuid, str)

    def test_project_data_connector_str(self, project_data_connector):
        string = str(project_data_connector)
        regex = r"^{.*\"id\": \"" + project_data_connector.id + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_project_data_connector_repr(self, project_data_connector):
        string = repr(project_data_connector)
        regex = (
            r"^ProjectDataConnector\( proximl , \*\*{.*'id': '"
            + project_data_connector.id
            + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_project_data_connector_bool(self, project_data_connector, mock_proximl):
        empty_project_data_connector = specimen.ProjectDataConnector(mock_proximl)
        assert bool(project_data_connector)
        assert not bool(empty_project_data_connector)



from pytest import fixture, mark


@fixture(scope="module")
@mark.xdist_group("project_resources")
async def project(proximl):
    project = await proximl.projects.create(
        name="New Project", copy_credentials=False, copy_secrets=False
    )
    yield project
    await project.remove()



import asyncio
import click
import logging
from os import devnull
from sys import stderr, stdout


from proximl.proximl import ProxiML


class ProxiMLRunner(object):
    def __init__(self):
        self._proximl_client = None

    @property
    def client(self) -> ProxiML:
        if self._proximl_client is None:
            try:
                self._proximl_client = ProxiML()
            except Exception as err:
                raise click.UsageError(err)
        return self._proximl_client

    async def _run(self, *tasks):
        return await asyncio.gather(*tasks)

    def run(self, *tasks):
        try:
            if len(tasks) == 1:
                return_value = asyncio.run(*tasks)
            else:
                return_value = asyncio.run(self._run(*tasks))
        except Exception as err:
            raise click.UsageError(err)
        return return_value


class Config(object):
    def __init__(self):
        self.stderr = stderr
        self.stdout = stdout
        self.proximl = ProxiMLRunner()


def search_by_id_name(term, list):
    found = None
    for item in list:
        if item.id == term:
            found = item
            break
    if None is found:
        for item in list:
            try:
                if item.name == term:
                    found = item
                    break
            except AttributeError:
                break
    return found


pass_config = click.make_pass_decorator(Config, ensure=True)


@click.group()
@click.version_option(package_name="proximl", prog_name="proxiML CLI and SDK")
@click.option(
    "--debug",
    "-d",
    is_flag=True,
    type=click.BOOL,
    default=False,
    help="Show debug output.",
)
@click.option(
    "--output-file",
    "-o",
    type=click.File("w"),
    default="-",
    help="Send output to file.",
)
@click.option(
    "--silent",
    "-s",
    is_flag=True,
    type=click.BOOL,
    default=False,
    help="Silence all output.",
)
@click.option(
    "--verbose",
    "-v",
    "verbosity",
    count=True,
    type=click.INT,
    default=0,
    help="Specify verbosity (repeat to increase).",
)
@pass_config
def cli(config, debug, output_file, silent, verbosity):
    """proxiML command-line interface."""
    config.stdout = output_file

    if debug or verbosity > 0:
        if silent:
            click.echo(
                "Ignoring silent flag when debug or verbosity is set.",
                file=config.stderr,
            )
        if verbosity == 1:
            verbosity = logging.INFO
        else:
            verbosity = logging.DEBUG
    elif silent:
        config.stderr = config.stdout = open(devnull, "w")
    else:
        verbosity = logging.WARNING

    if verbosity != logging.WARNING:  # default
        click.echo(
            f"Verbosity set to {logging.getLevelName(verbosity)}",
            file=config.stderr,
        )

    logging.basicConfig(
        format="%(asctime)s.%(msecs)03dZ  %(levelname)s  %(message)s",
        datefmt="%Y-%m-%dT%H:%M:%S",
        level=verbosity,
        stream=config.stderr,
        force=True,
    )


@cli.command()
@pass_config
def configure(config):
    active_project_id = config.proximl.client.active_project
    projects = config.proximl.run(config.proximl.client.projects.list())
    project_names = [project.name for project in projects]

    active_project = [
        project for project in projects if project.id == active_project_id
    ]

    active_project_name = active_project[0].name if len(active_project) else "UNSET"

    click.echo(f"Current Active Project: {active_project_name}")

    name = click.prompt(
        "Select Active Project:",
        type=click.Choice(project_names, case_sensitive=True),
        show_choices=True,
        default=active_project_name,
    )
    selected_project = [project for project in projects if project.name == name]
    config.proximl.client.set_active_project(selected_project[0].id)


from proximl.cli.connection import connection
from proximl.cli.dataset import dataset
from proximl.cli.model import model
from proximl.cli.checkpoint import checkpoint
from proximl.cli.volume import volume
from proximl.cli.environment import environment
from proximl.cli.gpu import gpu
from proximl.cli.job import job
from proximl.cli.project import project
from proximl.cli.cloudbender import cloudbender



from proximl.proximl import ProxiML
import asyncio


proximl_client = ProxiML()

# Create the dataset
dataset = asyncio.run(
    proximl_client.datasets.create(
        name="Example Dataset",
        source_type="aws",
        source_uri="s3://proximl-examples/data/cifar10",
    )
)

print(dataset)

# Watch the log output, attach will return when data transfer is complete
asyncio.run(dataset.attach())

# Create the job
job = asyncio.run(
    proximl_client.jobs.create(
        name="Example Training Job",
        type="training",
        gpu_types=["rtx2080ti", "rtx3090"],
        gpu_count=1,
        disk_size=10,
        workers=[
            "python training/image-classification/resnet_cifar.py --epochs 10 --optimizer adam --batch-size 128",
        ],
        data=dict(
            datasets=[dataset.id],
            output_uri="s3://proximl-examples/output/resnet_cifar10",
            output_type="aws",
        ),
        model=dict(
            source_type="git",
            source_uri="https://github.com/proxiML/examples.git",
        ),
    )
)
print(job)

# Watch the log output, attach will return when the training job stops
asyncio.run(job.attach())



import click
from webbrowser import open as browse
from proximl.cli import cli, pass_config, search_by_id_name


@cli.group()
@pass_config
def cloudbender(config):
    """proxiML CloudBender™ commands."""
    pass


from proximl.cli.cloudbender.provider import provider
from proximl.cli.cloudbender.region import region
from proximl.cli.cloudbender.node import node
from proximl.cli.cloudbender.device import device
from proximl.cli.cloudbender.datastore import datastore
from proximl.cli.cloudbender.data_connector import data_connector
from proximl.cli.cloudbender.service import service



import re
import json
import logging
from unittest.mock import AsyncMock, patch
from pytest import mark, fixture, raises
from aiohttp import WSMessage, WSMsgType

import proximl.cloudbender.services as specimen
from proximl.exceptions import (
    ApiError,
    SpecificationError,
    ProxiMLException,
)

pytestmark = [mark.sdk, mark.unit, mark.cloudbender, mark.services]


@fixture
def services(mock_proximl):
    yield specimen.Services(mock_proximl)


@fixture
def service(mock_proximl):
    yield specimen.Service(
        mock_proximl,
        provider_uuid="1",
        region_uuid="a",
        service_id="x",
        name="On-Prem Service",
        type="https",
        public=False,
        hostname="app1.proximl.cloud",
    )


class RegionsTests:
    @mark.asyncio
    async def test_get_service(
        self,
        services,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await services.get("1234", "5687", "91011")
        mock_proximl._query.assert_called_once_with(
            "/provider/1234/region/5687/service/91011", "GET", {}
        )

    @mark.asyncio
    async def test_list_services(
        self,
        services,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await services.list("1234", "5687")
        mock_proximl._query.assert_called_once_with(
            "/provider/1234/region/5687/service", "GET", {}
        )

    @mark.asyncio
    async def test_remove_service(
        self,
        services,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await services.remove("1234", "4567", "8910")
        mock_proximl._query.assert_called_once_with(
            "/provider/1234/region/4567/service/8910", "DELETE", {}
        )

    @mark.asyncio
    async def test_create_service(self, services, mock_proximl):
        requested_config = dict(
            provider_uuid="provider-id-1",
            region_uuid="region-id-1",
            name="On-Prem Service",
            type="https",
            public=False,
        )
        expected_payload = dict(
            name="On-Prem Service",
            type="https",
            public=False,
        )
        api_response = {
            "provider_uuid": "provider-id-1",
            "region_uuid": "region-id-1",
            "service_id": "service-id-1",
            "name": "On-Prem Service",
            "type": "https",
            "public": False,
            "hostname": "app1.proximl.cloud",
            "createdAt": "2020-12-31T23:59:59.000Z",
        }

        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await services.create(**requested_config)
        mock_proximl._query.assert_called_once_with(
            "/provider/provider-id-1/region/region-id-1/service",
            "POST",
            None,
            expected_payload,
        )
        assert response.id == "service-id-1"


class serviceTests:
    def test_service_properties(self, service):
        assert isinstance(service.id, str)
        assert isinstance(service.provider_uuid, str)
        assert isinstance(service.region_uuid, str)
        assert isinstance(service.public, bool)
        assert isinstance(service.name, str)
        assert isinstance(service.hostname, str)
        assert isinstance(service.type, str)

    def test_service_str(self, service):
        string = str(service)
        regex = r"^{.*\"service_id\": \"" + service.id + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_service_repr(self, service):
        string = repr(service)
        regex = r"^Service\( proximl , \*\*{.*'service_id': '" + service.id + r"'.*}\)$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_service_bool(self, service, mock_proximl):
        empty_service = specimen.Service(mock_proximl)
        assert bool(service)
        assert not bool(empty_service)

    @mark.asyncio
    async def test_service_remove(self, service, mock_proximl):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await service.remove()
        mock_proximl._query.assert_called_once_with(
            "/provider/1/region/a/service/x", "DELETE"
        )

    @mark.asyncio
    async def test_service_refresh(self, service, mock_proximl):
        api_response = {
            "provider_uuid": "provider-id-1",
            "region_uuid": "region-id-1",
            "service_id": "service-id-1",
            "name": "On-Prem Service",
            "type": "https",
            "public": False,
            "hostname": "app1.proximl.cloud",
            "createdAt": "2020-12-31T23:59:59.000Z",
        }
        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await service.refresh()
        mock_proximl._query.assert_called_once_with(
            f"/provider/1/region/a/service/x", "GET"
        )
        assert service.id == "service-id-1"
        assert response.id == "service-id-1"



import re
import sys
import asyncio
from pytest import mark, fixture

pytestmark = [mark.sdk, mark.integration, mark.projects]


@mark.create
@mark.asyncio
@mark.xdist_group("project_resources")
class ProjectDataConnectorsTests:
    @fixture(scope="class")
    async def project_data_connector(self, project):
        data_connectors = await project.data_connectors.list()
        yield data_connectors[0]

    async def test_list_project_data_connectors(self, project, project_data_connector):
        data_connectors = await project.data_connectors.list()
        assert len(data_connectors) > 0

    async def test_project_data_connector_properties(
        self, project, project_data_connector
    ):
        assert isinstance(project_data_connector.project_uuid, str)
        assert isinstance(project_data_connector.type, str)
        assert isinstance(project_data_connector.name, str)
        assert isinstance(project_data_connector.region_uuid, str)
        assert project.id == project_data_connector.project_uuid

    async def test_project_data_connector_str(self, project_data_connector):
        string = str(project_data_connector)
        regex = r"^{.*\"name\": \"" + project_data_connector.name + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    async def test_project_data_connector_repr(self, project_data_connector):
        string = repr(project_data_connector)
        regex = (
            r"^ProjectDataConnector\( proximl , \*\*{.*'name': '"
            + project_data_connector.name
            + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)



import re
import sys
import asyncio
from pytest import mark, fixture

pytestmark = [mark.sdk, mark.integration, mark.projects]


@mark.create
@mark.asyncio
@mark.xdist_group("project_resources")
class ProjectCredentialsTests:
    @fixture(scope="class")
    async def project_credential(self, project):
        project_credential = await project.credentials.put(
            type="aws", key_id="ASFHALKF", secret="IUHKLHKAHF"
        )
        yield project_credential
        await project.credentials.remove(type="aws")

    async def test_list_project_credentials(self, project, project_credential):
        credentials = await project.credentials.list()
        assert len(credentials) > 0

    async def test_project_credential_properties(self, project, project_credential):
        assert isinstance(project_credential.project_uuid, str)
        assert isinstance(project_credential.type, str)
        assert isinstance(project_credential.key_id, str)
        assert project_credential.type == "aws"
        assert project.id == project_credential.project_uuid

    async def test_project_credential_str(self, project_credential):
        string = str(project_credential)
        regex = r"^{.*\"type\": \"" + project_credential.type + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    async def test_project_credential_repr(self, project_credential):
        string = repr(project_credential)
        regex = (
            r"^ProjectCredential\( proximl , \*\*{.*'type': '"
            + project_credential.type
            + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)



import re
import json
import logging
from unittest.mock import AsyncMock, patch
from pytest import mark, fixture, raises
from aiohttp import WSMessage, WSMsgType

import proximl.checkpoints as specimen
from proximl.exceptions import (
    ApiError,
    CheckpointError,
    SpecificationError,
    ProxiMLException,
)

pytestmark = [mark.sdk, mark.unit, mark.checkpoints]


@fixture
def checkpoints(mock_proximl):
    yield specimen.Checkpoints(mock_proximl)


@fixture
def checkpoint(mock_proximl):
    yield specimen.Checkpoint(
        mock_proximl,
        checkpoint_uuid="1",
        project_uuid="proj-id-1",
        name="first one",
        status="downloading",
        size=100000,
        createdAt="2020-12-31T23:59:59.000Z",
    )


class CheckpointsTests:
    @mark.asyncio
    async def test_get_checkpoint(
        self,
        checkpoints,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await checkpoints.get("1234")
        mock_proximl._query.assert_called_once_with("/checkpoint/1234", "GET", dict())

    @mark.asyncio
    async def test_list_checkpoints(
        self,
        checkpoints,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await checkpoints.list()
        mock_proximl._query.assert_called_once_with("/checkpoint", "GET", dict())

    @mark.asyncio
    async def test_remove_checkpoint(
        self,
        checkpoints,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await checkpoints.remove("4567")
        mock_proximl._query.assert_called_once_with(
            "/checkpoint/4567", "DELETE", dict(force=True)
        )

    @mark.asyncio
    async def test_create_checkpoint_simple(self, checkpoints, mock_proximl):
        requested_config = dict(
            name="new checkpoint",
            source_type="aws",
            source_uri="s3://proximl-examples/checkpoints/resnet50",
        )
        expected_payload = dict(
            project_uuid="proj-id-1",
            name="new checkpoint",
            source_type="aws",
            source_uri="s3://proximl-examples/checkpoints/resnet50",
            type="evefs",
        )
        api_response = {
            "project_uuid": "cus-id-1",
            "checkpoint_uuid": "checkpoint-id-1",
            "name": "new checkpoint",
            "type": "evefs",
            "status": "new",
            "source_type": "aws",
            "source_uri": "s3://proximl-examples/checkpoints/resnet50",
            "createdAt": "2020-12-20T16:46:23.909Z",
        }

        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await checkpoints.create(**requested_config)
        mock_proximl._query.assert_called_once_with(
            "/checkpoint", "POST", None, expected_payload
        )
        assert response.id == "checkpoint-id-1"


class CheckpointTests:
    def test_checkpoint_properties(self, checkpoint):
        assert isinstance(checkpoint.id, str)
        assert isinstance(checkpoint.status, str)
        assert isinstance(checkpoint.name, str)
        assert isinstance(checkpoint.size, int)

    def test_checkpoint_str(self, checkpoint):
        string = str(checkpoint)
        regex = r"^{.*\"checkpoint_uuid\": \"" + checkpoint.id + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_checkpoint_repr(self, checkpoint):
        string = repr(checkpoint)
        regex = (
            r"^Checkpoint\( proximl , \*\*{.*'checkpoint_uuid': '"
            + checkpoint.id
            + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_checkpoint_bool(self, checkpoint, mock_proximl):
        empty_checkpoint = specimen.Checkpoint(mock_proximl)
        assert bool(checkpoint)
        assert not bool(empty_checkpoint)

    @mark.asyncio
    async def test_checkpoint_get_log_url(self, checkpoint, mock_proximl):
        api_response = (
            "https://trainml-jobs-dev.s3.us-east-2.amazonaws.com/1/logs/first_one.zip"
        )
        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await checkpoint.get_log_url()
        mock_proximl._query.assert_called_once_with(
            "/checkpoint/1/logs", "GET", dict(project_uuid="proj-id-1")
        )
        assert response == api_response

    @mark.asyncio
    async def test_checkpoint_get_details(self, checkpoint, mock_proximl):
        api_response = {
            "type": "directory",
            "name": "/",
            "count": "8",
            "size": "177M",
            "contents": [],
        }
        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await checkpoint.get_details()
        mock_proximl._query.assert_called_once_with(
            "/checkpoint/1/details", "GET", dict(project_uuid="proj-id-1")
        )
        assert response == api_response

    @mark.asyncio
    async def test_checkpoint_get_connection_utility_url(
        self, checkpoint, mock_proximl
    ):
        api_response = (
            "https://trainml-jobs-dev.s3.us-east-2.amazonaws.com/1/vpn/first_one.zip"
        )
        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await checkpoint.get_connection_utility_url()
        mock_proximl._query.assert_called_once_with(
            "/checkpoint/1/download", "GET", dict(project_uuid="proj-id-1")
        )
        assert response == api_response

    def test_checkpoint_get_connection_details_no_vpn(self, checkpoint):
        details = checkpoint.get_connection_details()
        expected_details = dict()
        assert details == expected_details

    def test_checkpoint_get_connection_details_local_data(self, mock_proximl):
        checkpoint = specimen.Checkpoint(
            mock_proximl,
            checkpoint_uuid="1",
            project_uuid="a",
            name="first one",
            status="new",
            size=100000,
            createdAt="2020-12-31T23:59:59.000Z",
            source_type="local",
            source_uri="~/tensorflow-example",
            vpn={
                "status": "new",
                "cidr": "10.106.171.0/24",
                "client": {
                    "port": "36017",
                    "id": "cus-id-1",
                    "address": "10.106.171.253",
                    "ssh_port": 46600,
                },
                "net_prefix_type_id": 1,
            },
        )
        details = checkpoint.get_connection_details()
        expected_details = dict(
            project_uuid="a",
            entity_type="checkpoint",
            cidr="10.106.171.0/24",
            ssh_port=46600,
            input_path="~/tensorflow-example",
            output_path=None,
        )
        assert details == expected_details

    @mark.asyncio
    async def test_checkpoint_connect(self, checkpoint, mock_proximl):
        with patch(
            "proximl.checkpoints.Connection",
            autospec=True,
        ) as mock_connection:
            connection = mock_connection.return_value
            connection.status = "connected"
            resp = await checkpoint.connect()
            connection.start.assert_called_once()
            assert resp == "connected"

    @mark.asyncio
    async def test_checkpoint_disconnect(self, checkpoint, mock_proximl):
        with patch(
            "proximl.checkpoints.Connection",
            autospec=True,
        ) as mock_connection:
            connection = mock_connection.return_value
            connection.status = "removed"
            resp = await checkpoint.disconnect()
            connection.stop.assert_called_once()
            assert resp == "removed"

    @mark.asyncio
    async def test_checkpoint_remove(self, checkpoint, mock_proximl):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await checkpoint.remove()
        mock_proximl._query.assert_called_once_with(
            "/checkpoint/1",
            "DELETE",
            dict(project_uuid="proj-id-1", force=False),
        )

    def test_checkpoint_default_ws_msg_handler(self, checkpoint, capsys):
        data = {
            "msg": "download: s3://proximl-examples/data/cifar10/data_batch_2.bin to ./data_batch_2.bin\n",
            "time": 1613079345318,
            "type": "subscription",
            "stream": "worker-id-1",
            "job_worker_uuid": "worker-id-1",
        }

        handler = checkpoint._get_msg_handler(None)
        handler(data)
        captured = capsys.readouterr()
        assert (
            captured.out
            == "02/11/2021, 15:35:45: download: s3://proximl-examples/data/cifar10/data_batch_2.bin to ./data_batch_2.bin\n"
        )

    def test_checkpoint_custom_ws_msg_handler(self, checkpoint, capsys):
        def custom_handler(msg):
            print(msg.get("stream"))

        data = {
            "msg": "download: s3://proximl-examples/data/cifar10/data_batch_2.bin to ./data_batch_2.bin\n",
            "time": 1613079345318,
            "type": "subscription",
            "stream": "worker-id-1",
            "job_worker_uuid": "worker-id-1",
        }

        handler = checkpoint._get_msg_handler(custom_handler)
        handler(data)
        captured = capsys.readouterr()
        assert captured.out == "worker-id-1\n"

    @mark.asyncio
    async def test_checkpoint_attach(self, checkpoint, mock_proximl):
        api_response = None
        mock_proximl._ws_subscribe = AsyncMock(return_value=api_response)
        refresh_response = {
            "customer_uuid": "cus-id-1",
            "checkpoint_uuid": "data-id-1",
            "name": "new checkpoint",
            "status": "downloading",
            "source_type": "aws",
            "source_uri": "s3://proximl-examples/data/cifar10",
            "createdAt": "2020-12-20T16:46:23.909Z",
        }
        checkpoint.refresh = AsyncMock(return_value=refresh_response)
        await checkpoint.attach()
        mock_proximl._ws_subscribe.assert_called_once()

    @mark.asyncio
    async def test_checkpoint_attach_immediate_return(self, mock_proximl):
        checkpoint = specimen.Checkpoint(
            mock_proximl,
            checkpoint_uuid="1",
            name="first one",
            status="ready",
            createdAt="2020-12-31T23:59:59.000Z",
        )
        api_response = None
        mock_proximl._ws_subscribe = AsyncMock(return_value=api_response)
        refresh_response = {
            "customer_uuid": "cus-id-1",
            "checkpoint_uuid": "1",
            "name": "new checkpoint",
            "status": "ready",
            "createdAt": "2020-12-20T16:46:23.909Z",
        }
        checkpoint.refresh = AsyncMock(return_value=refresh_response)
        await checkpoint.attach()
        mock_proximl._ws_subscribe.assert_not_called()

    @mark.asyncio
    async def test_checkpoint_refresh(self, checkpoint, mock_proximl):
        api_response = {
            "customer_uuid": "cus-id-1",
            "checkpoint_uuid": "data-id-1",
            "name": "new checkpoint",
            "status": "ready",
            "source_type": "aws",
            "source_uri": "s3://proximl-examples/data/cifar10",
            "createdAt": "2020-12-20T16:46:23.909Z",
        }
        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await checkpoint.refresh()
        mock_proximl._query.assert_called_once_with(
            f"/checkpoint/1", "GET", dict(project_uuid="proj-id-1")
        )
        assert checkpoint.id == "data-id-1"
        assert response.id == "data-id-1"

    @mark.asyncio
    async def test_checkpoint_wait_for_successful(self, checkpoint, mock_proximl):
        api_response = {
            "customer_uuid": "cus-id-1",
            "checkpoint_uuid": "data-id-1",
            "name": "new checkpoint",
            "status": "ready",
            "source_type": "aws",
            "source_uri": "s3://proximl-examples/data/cifar10",
            "createdAt": "2020-12-20T16:46:23.909Z",
        }
        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await checkpoint.wait_for("ready")
        mock_proximl._query.assert_called_once_with(
            f"/checkpoint/1", "GET", dict(project_uuid="proj-id-1")
        )
        assert checkpoint.id == "data-id-1"
        assert response.id == "data-id-1"

    @mark.asyncio
    async def test_checkpoint_wait_for_current_status(self, mock_proximl):
        checkpoint = specimen.Checkpoint(
            mock_proximl,
            checkpoint_uuid="1",
            name="first one",
            status="ready",
            createdAt="2020-12-31T23:59:59.000Z",
        )
        api_response = None
        mock_proximl._query = AsyncMock(return_value=api_response)
        await checkpoint.wait_for("ready")
        mock_proximl._query.assert_not_called()

    @mark.asyncio
    async def test_checkpoint_wait_for_incorrect_status(self, checkpoint, mock_proximl):
        api_response = None
        mock_proximl._query = AsyncMock(return_value=api_response)
        with raises(SpecificationError):
            await checkpoint.wait_for("stopped")
        mock_proximl._query.assert_not_called()

    @mark.asyncio
    async def test_checkpoint_wait_for_with_delay(self, checkpoint, mock_proximl):
        api_response_initial = dict(
            checkpoint_uuid="1",
            name="first one",
            status="new",
            createdAt="2020-12-31T23:59:59.000Z",
        )
        api_response_final = dict(
            checkpoint_uuid="1",
            name="first one",
            status="ready",
            createdAt="2020-12-31T23:59:59.000Z",
        )
        mock_proximl._query = AsyncMock()
        mock_proximl._query.side_effect = [
            api_response_initial,
            api_response_initial,
            api_response_final,
        ]
        response = await checkpoint.wait_for("ready")
        assert checkpoint.status == "ready"
        assert response.status == "ready"

    @mark.asyncio
    async def test_checkpoint_wait_for_timeout(self, checkpoint, mock_proximl):
        api_response = dict(
            checkpoint_uuid="1",
            name="first one",
            status="downloading",
            createdAt="2020-12-31T23:59:59.000Z",
        )
        mock_proximl._query = AsyncMock(return_value=api_response)
        with raises(ProxiMLException):
            await checkpoint.wait_for("ready", 10)
        mock_proximl._query.assert_called()

    @mark.asyncio
    async def test_checkpoint_wait_for_checkpoint_failed(
        self, checkpoint, mock_proximl
    ):
        api_response = dict(
            checkpoint_uuid="1",
            name="first one",
            status="failed",
            createdAt="2020-12-31T23:59:59.000Z",
        )
        mock_proximl._query = AsyncMock(return_value=api_response)
        with raises(CheckpointError):
            await checkpoint.wait_for("ready")
        mock_proximl._query.assert_called()

    @mark.asyncio
    async def test_checkpoint_wait_for_archived_succeeded(
        self, checkpoint, mock_proximl
    ):
        mock_proximl._query = AsyncMock(
            side_effect=ApiError(404, dict(errorMessage="Checkpoint Not Found"))
        )
        await checkpoint.wait_for("archived")
        mock_proximl._query.assert_called()

    @mark.asyncio
    async def test_checkpoint_wait_for_unexpected_api_error(
        self, checkpoint, mock_proximl
    ):
        mock_proximl._query = AsyncMock(
            side_effect=ApiError(404, dict(errorMessage="Checkpoint Not Found"))
        )
        with raises(ApiError):
            await checkpoint.wait_for("ready")
        mock_proximl._query.assert_called()



import re
import json
import click
from unittest.mock import AsyncMock, patch, create_autospec
from pytest import mark, fixture, raises

pytestmark = [mark.cli, mark.unit, mark.projects]

from proximl.cli import project as specimen


def test_list(runner, mock_projects):
    with patch("proximl.cli.ProxiML", new=AsyncMock) as mock_proximl:
        mock_proximl.projects = AsyncMock()
        mock_proximl.projects.list = AsyncMock(return_value=mock_projects)
        result = runner.invoke(specimen, ["list"])
        print(result)
        assert result.exit_code == 0
        mock_proximl.projects.list.assert_called_once()



import json
import logging
from typing import Literal

class ProjectMembers(object):
    def __init__(self, proximl, project_id):
        self.proximl = proximl
        self.project_id = project_id

    async def list(self, **kwargs):
        resp = await self.proximl._query(
            f"/project/{self.project_id}/access", "GET", kwargs
        )
        members = [ProjectMember(self.proximl, **member) for member in resp]
        return members
    
    async def add(self, email: str, job: Literal["all", "read"], dataset: Literal["all", "read"], model: Literal["all", "read"], checkpoint: Literal["all", "read"], volume: Literal["all", "read"],  **kwargs):
        data = dict(
            email=email,
            job=job,
            dataset=dataset,
            model=model,
            checkpoint=checkpoint,
            volume=volume,
        )
        payload = {k: v for k, v in data.items() if v is not None}
        resp = await self.proximl._query(
            f"/project/{self.project_id}/access", "POST",kwargs, payload)
        member = ProjectMember(self.proximl, **resp)
        logging.info(f"Added Project Member {email} to project {self.project_id}")
        return member

    
    async def remove(self, email, **kwargs):
        await self.proximl._query(
            f"/project/{self.project_id}/access", "DELETE", dict(**kwargs, email=email)
        )


class ProjectMember:
    def __init__(self, proximl, **kwargs):
        self.proximl = proximl
        self._entity = kwargs
        self._id = self._entity.get("email")
        self._project_uuid = self._entity.get("project_uuid")
        self._owner = self._entity.get("owner")
        self._job = self._entity.get("job")
        self._dataset = self._entity.get("dataset")
        self._model = self._entity.get("model")
        self._checkpoint = self._entity.get("checkpoint")
        self._volume = self._entity.get("volume")

    @property
    def id(self) -> str:
        return self._id

    @property
    def project_uuid(self) -> str:
        return self._project_uuid
    
    @property
    def email(self) -> str:
        return self._id

    @property
    def owner(self) -> bool:
        return self._owner

    @property
    def job(self) -> str:
        return self._job

    @property
    def dataset(self) -> str:
        return self._dataset
    
    @property
    def model(self) -> str:
        return self._model
    
    @property
    def checkpoint(self) -> str:
        return self._checkpoint
    
    @property
    def volume(self) -> str:
        return self._volume

    def __str__(self):
        return json.dumps({k: v for k, v in self._entity.items()})

    def __repr__(self):
        return f"ProjectMember( proximl , **{self._entity.__repr__()})"

    def __bool__(self):
        return bool(self._id)





import re
import json
import logging
from unittest.mock import AsyncMock, patch
from pytest import mark, fixture, raises
from aiohttp import WSMessage, WSMsgType

import proximl.cloudbender.regions as specimen
from proximl.exceptions import (
    ApiError,
    SpecificationError,
    ProxiMLException,
)

pytestmark = [mark.sdk, mark.unit, mark.cloudbender, mark.regions]


@fixture
def regions(mock_proximl):
    yield specimen.Regions(mock_proximl)


@fixture
def region(mock_proximl):
    yield specimen.Region(
        mock_proximl,
        provider_uuid="1",
        region_uuid="a",
        provider_type="physical",
        name="Physical Region 1",
        createdAt="2020-12-31T23:59:59.000Z",
        status="healthy",
    )


class RegionsTests:
    @mark.asyncio
    async def test_get_region(
        self,
        regions,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await regions.get("1234", "5687")
        mock_proximl._query.assert_called_once_with(
            "/provider/1234/region/5687", "GET", {}
        )

    @mark.asyncio
    async def test_list_regions(
        self,
        regions,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await regions.list("1234")
        mock_proximl._query.assert_called_once_with(
            "/provider/1234/region", "GET", {}
        )

    @mark.asyncio
    async def test_remove_region(
        self,
        regions,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await regions.remove("1234", "4567")
        mock_proximl._query.assert_called_once_with(
            "/provider/1234/region/4567", "DELETE", {}
        )

    @mark.asyncio
    async def test_create_region(self, regions, mock_proximl):
        requested_config = dict(
            provider_uuid="provider-id-1",
            name="phys-region",
            public=False,
            storage=dict(mode="local"),
        )
        expected_payload = dict(
            name="phys-region", public=False, storage=dict(mode="local")
        )
        api_response = {
            "provider_uuid": "provider-id-1",
            "region_uuid": "region-id-1",
            "provider_type": "physical",
            "name": "phys-region",
            "status": "new",
            "createdAt": "2020-12-31T23:59:59.000Z",
        }

        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await regions.create(**requested_config)
        mock_proximl._query.assert_called_once_with(
            "/provider/provider-id-1/region", "POST", None, expected_payload
        )
        assert response.id == "region-id-1"


class regionTests:
    def test_region_properties(self, region):
        assert isinstance(region.id, str)
        assert isinstance(region.provider_uuid, str)
        assert isinstance(region.type, str)
        assert isinstance(region.name, str)
        assert isinstance(region.status, str)

    def test_region_str(self, region):
        string = str(region)
        regex = r"^{.*\"region_uuid\": \"" + region.id + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_region_repr(self, region):
        string = repr(region)
        regex = (
            r"^Region\( proximl , \*\*{.*'region_uuid': '"
            + region.id
            + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_region_bool(self, region, mock_proximl):
        empty_region = specimen.Region(mock_proximl)
        assert bool(region)
        assert not bool(empty_region)

    @mark.asyncio
    async def test_region_remove(self, region, mock_proximl):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await region.remove()
        mock_proximl._query.assert_called_once_with(
            "/provider/1/region/a", "DELETE"
        )

    @mark.asyncio
    async def test_region_refresh(self, region, mock_proximl):
        api_response = {
            "provider_uuid": "provider-id-1",
            "region_uuid": "region-id-1",
            "provider_type": "physical",
            "name": "phys-region",
            "status": "new",
            "createdAt": "2020-12-31T23:59:59.000Z",
        }
        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await region.refresh()
        mock_proximl._query.assert_called_once_with(
            f"/provider/1/region/a", "GET"
        )
        assert region.id == "region-id-1"
        assert response.id == "region-id-1"

    @mark.asyncio
    async def test_region_stage_dataset(self, region, mock_proximl):
        api_response = dict()
        expected_payload = dict(
            project_uuid="proj-id-1",
            dataset_uuid="dataset-id-1",
        )
        mock_proximl._query = AsyncMock(return_value=api_response)
        await region.add_dataset("proj-id-1", "dataset-id-1")
        mock_proximl._query.assert_called_once_with(
            "/provider/1/region/a/dataset", "POST", None, expected_payload
        )

    @mark.asyncio
    async def test_region_stage_model(self, region, mock_proximl):
        api_response = dict()
        expected_payload = dict(
            project_uuid="proj-id-1",
            model_uuid="model-id-1",
        )
        mock_proximl._query = AsyncMock(return_value=api_response)
        await region.add_model("proj-id-1", "model-id-1")
        mock_proximl._query.assert_called_once_with(
            "/provider/1/region/a/model", "POST", None, expected_payload
        )

    @mark.asyncio
    async def test_region_stage_checkpoint(self, region, mock_proximl):
        api_response = dict()
        expected_payload = dict(
            project_uuid="proj-id-1",
            checkpoint_uuid="checkpoint-id-1",
        )
        mock_proximl._query = AsyncMock(return_value=api_response)
        await region.add_checkpoint("proj-id-1", "checkpoint-id-1")
        mock_proximl._query.assert_called_once_with(
            "/provider/1/region/a/checkpoint", "POST", None, expected_payload
        )



import click
import os
import json
import base64
from pathlib import Path
from proximl.cli import pass_config
from proximl.cli.project import project


@project.group()
@pass_config
def credential(config):
    """proxiML project credential commands."""
    pass


@credential.command()
@pass_config
def list(config):
    """List credentials."""
    data = [
        ["TYPE", "KEY ID", "UPDATED AT"],
        [
            "-" * 80,
            "-" * 80,
            "-" * 80,
        ],
    ]
    project = config.proximl.run(config.proximl.client.projects.get_current())
    credentials = config.proximl.run(project.credentials.list())

    for credential in credentials:
        data.append(
            [
                credential.type,
                credential.key_id,
                credential.updated_at.isoformat(timespec="seconds"),
            ]
        )

    for row in data:
        click.echo(
            "{: >13.11} {: >37.35} {: >28.26}" "".format(*row),
            file=config.stdout,
        )


@credential.command()
@click.argument(
    "type",
    type=click.Choice(
        [
            "aws",
            "azure",
            "docker",
            "gcp",
            "huggingface",
            "kaggle",
            "ngc",
            "wasabi",
        ],
        case_sensitive=False,
    ),
)
@pass_config
def put(config, type):
    """
    Set a credential.

    A credential is uploaded.
    """
    project = config.proximl.run(config.proximl.client.projects.get_current())

    tenant = None

    if type in ["aws", "wasabi"]:
        credential_id = click.prompt(
            "Enter the credential ID", type=str, hide_input=False
        )
        secret = click.prompt("Enter the secret credential", type=str, hide_input=True)
    elif type == "azure":
        credential_id = click.prompt(
            "Enter the Application (client) ID", type=str, hide_input=False
        )
        tenant = click.prompt(
            "Enter the Directory (tenant) ley", type=str, hide_input=False
        )
        secret = click.prompt("Enter the client secret", type=str, hide_input=True)
    elif type in ["docker", "huggingface"]:
        credential_id = click.prompt("Enter the username", type=str, hide_input=False)
        secret = click.prompt("Enter the access token", type=str, hide_input=True)
    elif type in ["gcp", "kaggle"]:
        file_name = click.prompt(
            "Enter the path of the credentials file",
            type=click.Path(
                exists=True, file_okay=True, dir_okay=False, resolve_path=True
            ),
            hide_input=False,
        )
        credential_id = os.path.basename(file_name)
        with open(file_name) as f:
            secret = json.load(f)
        secret = json.dumps(secret)
    elif type == "ngc":
        credential_id = "$oauthtoken"
        secret = click.prompt("Enter the access token", type=str, hide_input=True)
    else:
        raise click.UsageError("Unsupported credential type")

    return config.proximl.run(
        project.credentials.put(
            type=type, credential_id=credential_id, secret=secret, tenant=tenant
        )
    )


@credential.command()
@click.argument("name", type=click.STRING)
@pass_config
def remove(config, name):
    """
    Remove a credential.


    """
    project = config.proximl.run(config.proximl.client.projects.get_current())

    return config.proximl.run(project.credential.remove(name))



import re
import sys
import asyncio
from pytest import mark, fixture

pytestmark = [mark.sdk, mark.integration, mark.projects]


@mark.create
@mark.asyncio
@mark.xdist_group("project_resources")
class GetProjectsTests:
    async def test_get_projects(self, proximl):
        projects = await proximl.projects.list()
        assert len(projects) > 0

    async def test_get_project(self, proximl, project):
        response = await proximl.projects.get(project.id)
        assert response.id == project.id

    async def test_project_properties(self, project):
        assert isinstance(project.id, str)
        assert isinstance(project.name, str)
        assert isinstance(project.owner_name, str)
        assert isinstance(project.is_owner, bool)
        assert project.name == "New Project"
        assert project.is_owner

    async def test_project_str(self, project):
        string = str(project)
        regex = r"^{.*\"id\": \"" + project.id + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    async def test_project_repr(self, project):
        string = repr(project)
        regex = r"^Project\( proximl , \*\*{.*'id': '" + project.id + r"'.*}\)$"
        assert isinstance(string, str)
        assert re.match(regex, string)



import re
import json
import logging
from unittest.mock import AsyncMock, patch
from pytest import mark, fixture, raises
from aiohttp import WSMessage, WSMsgType

import proximl.projects.members as specimen
from proximl.exceptions import (
    ApiError,
    SpecificationError,
    ProxiMLException,
)

pytestmark = [mark.sdk, mark.unit, mark.projects]


@fixture
def project_members(mock_proximl):
    yield specimen.ProjectMembers(mock_proximl, project_id="1")


@fixture
def project_member(mock_proximl):
    yield specimen.ProjectMember(
        mock_proximl,
        id="owner@gmail.com",
        email="owner@gmail.com",
        project_uuid="proj-id-1",
       owner= True,
                job= "all",
                dataset= "all",
                model= "all",
                checkpoint="all",
                volume= "all"
    )


class ProjectMembersTests:
    @mark.asyncio
    async def test_project_members_list(self, project_members, mock_proximl):
        api_response = [
            {
                "project_uuid": "proj-id-1",
                "email": "owner@gmail.com",
                "createdAt": "2024-09-04T00:42:39.529Z",
                "updatedAt": "2024-09-04T00:42:39.529Z",
                "owner": True,
                "job": "all",
                "dataset": "all",
                "model": "all",
                "checkpoint": "all",
                "volume": "all"
            },
            {
                "project_uuid": "proj-id-1",
                "email": "non-owner@gmail.com",
                "createdAt": "2024-09-04T00:42:39.529Z",
                "updatedAt": "2024-09-04T00:42:39.529Z",
                "owner": False,
                "job": "all",
                "dataset": "all",
                "model": "all",
                "checkpoint": "read",
                "volume": "read"
            },
        ]
        mock_proximl._query = AsyncMock(return_value=api_response)
        resp = await project_members.list()
        mock_proximl._query.assert_called_once_with(
            "/project/1/access", "GET", dict()
        )
        assert len(resp) == 2


class ProjectMemberTests:
    def test_project_member_properties(self, project_member):
        assert isinstance(project_member.id, str)
        assert isinstance(project_member.email, str)
        assert isinstance(project_member.project_uuid, str)
        assert isinstance(project_member.owner, bool)
        assert isinstance(project_member.job, str)
        assert isinstance(project_member.dataset, str)
        assert isinstance(project_member.model, str)
        assert isinstance(project_member.checkpoint, str)
        assert isinstance(project_member.volume, str)

    def test_project_member_str(self, project_member):
        string = str(project_member)
        regex = r"^{.*\"id\": \"" + project_member.id + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_project_member_repr(self, project_member):
        string = repr(project_member)
        regex = (
            r"^ProjectMember\( proximl , \*\*{.*'id': '"
            + project_member.id
            + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_project_member_bool(self, project_member, mock_proximl):
        empty_project_member = specimen.ProjectMember(mock_proximl)
        assert bool(project_member)
        assert not bool(empty_project_member)



from proximl.proximl import ProxiML
import asyncio


proximl_client = ProxiML()

# Create the dataset
dataset = asyncio.run(
    proximl_client.datasets.create(
        name="Example Dataset",
        source_type="aws",
        source_uri="s3://proximl-examples/data/cifar10",
    )
)

print(dataset)

# Watch the log output, attach will return when data transfer is complete
asyncio.run(dataset.attach())

# Create the job
training_job = asyncio.run(
    proximl_client.jobs.create(
        name="Example Training Job",
        type="training",
        gpu_types=["rtx2080ti", "rtx3090"],
        gpu_count=1,
        disk_size=10,
        workers=[
            "python training/image-classification/resnet_cifar.py --epochs 10 --optimizer adam --batch-size 128",
        ],
        data=dict(
            datasets=[dataset.id], output_type="proximl", output_uri="model"
        ),
        model=dict(
            source_type="git",
            source_uri="https://github.com/proxiML/examples.git",
        ),
    )
)
print(training_job)

# Watch the log output, attach will return when the training job stops
asyncio.run(training_job.attach())

# Get the trained model id from the workers
training_job = asyncio.run(training_job.refresh())

model = asyncio.run(
    proximl_client.models.get(training_job.workers[0].get("output_uuid"))
)

# Ensure the model is ready to use
asyncio.run(model.wait_for("ready"))

# Use the model in an inference job on new data
inference_job = asyncio.run(
    proximl_client.jobs.create(
        name="Example Inference Job",
        type="inference",
        gpu_types=["rtx2080ti", "rtx3090"],
        gpu_count=1,
        disk_size=10,
        workers=[
            "python training/image-classification/resnet_cifar.py",
        ],
        data=dict(
            input_type="aws",
            input_uri="s3://proximl-examples/data/new_data",
            output_type="aws",
            output_uri="s3://proximl-examples/output/model_predictions",
        ),
        model=dict(source_type="proximl", source_uri=model.id),
    )
)
print(inference_job)

# Watch the log output, attach will return when the training job stops
asyncio.run(inference_job.attach())

# (Optional) Cleanup
asyncio.gather(
    training_job.remove(),
    inference_job.remove(),
    model.remove(),
    dataset.remove(),
)



import re
import json
import click
from unittest.mock import AsyncMock, patch, create_autospec
from pytest import mark, fixture, raises

pytestmark = [mark.cli, mark.unit, mark.projects]

from proximl.cli.project import service as specimen
from proximl.projects import (
    Project,
)


def test_list_services(runner, mock_project_services):
    with patch("proximl.cli.ProxiML", new=AsyncMock) as mock_proximl:
        project = create_autospec(Project)
        mock_proximl.projects = AsyncMock()
        mock_proximl.projects.get = AsyncMock(return_value=project)
        mock_proximl.projects.get_current = AsyncMock(return_value=project)
        project.services = AsyncMock()
        project.services.list = AsyncMock(return_value=mock_project_services)
        result = runner.invoke(specimen, ["list"])
        print(result)
        assert result.exit_code == 0
        project.services.list.assert_called_once()



import click
from proximl.cli import cli, pass_config, search_by_id_name
from proximl.cli.cloudbender import cloudbender


@cloudbender.group()
@pass_config
def service(config):
    """proxiML CloudBender service commands."""
    pass


@service.command()
@click.option(
    "--provider",
    "-p",
    type=click.STRING,
    required=True,
    help="The provider ID of the region.",
)
@click.option(
    "--region",
    "-r",
    type=click.STRING,
    required=True,
    help="The region ID to list services for.",
)
@pass_config
def list(config, provider, region):
    """List services."""
    data = [
        ["ID", "NAME", "HOSTNAME"],
        [
            "-" * 80,
            "-" * 80,
            "-" * 80,
        ],
    ]

    services = config.proximl.run(
        config.proximl.client.cloudbender.services.list(
            provider_uuid=provider, region_uuid=region
        )
    )

    for service in services:
        data.append(
            [
                service.id,
                service.name,
                service.hostname,
            ]
        )

    for row in data:
        click.echo(
            "{: >25.24} {: >29.28} {: >40.39}" "".format(*row),
            file=config.stdout,
        )


@service.command()
@click.option(
    "--provider",
    "-p",
    type=click.STRING,
    required=True,
    help="The provider ID of the region.",
)
@click.option(
    "--region",
    "-r",
    type=click.STRING,
    required=True,
    help="The region ID to create the service in.",
)
@click.option(
    "--type",
    "-t",
    type=click.Choice(
        [
            "https",
            "tcp",
            "udp",
        ],
    ),
    required=True,
    help="The type of regional service.",
)
@click.option(
    "--public/--no-public",
    default=True,
    show_default=True,
    help="Service should be accessible from the public internet.",
)
@click.argument("name", type=click.STRING, required=True)
@pass_config
def create(config, provider, region, type, public, name):
    """
    Creates a service.
    """
    return config.proximl.run(
        config.proximl.client.cloudbender.services.create(
            provider_uuid=provider,
            region_uuid=region,
            name=name,
            type=type,
            public=public,
        )
    )


@service.command()
@click.option(
    "--provider",
    "-p",
    type=click.STRING,
    required=True,
    help="The provider ID of the region.",
)
@click.option(
    "--region",
    "-r",
    type=click.STRING,
    required=True,
    help="The region ID to remove the service from.",
)
@click.argument("service", type=click.STRING)
@pass_config
def remove(config, provider, region, service):
    """
    Remove a service.

    RESERVATION may be specified by name or ID, but ID is preferred.
    """
    services = config.proximl.run(
        config.proximl.client.cloudbender.services.list(
            provider_uuid=provider, region_uuid=region
        )
    )

    found = search_by_id_name(service, services)
    if None is found:
        raise click.UsageError("Cannot find specified service.")

    return config.proximl.run(found.remove())



import re
import sys
import tempfile
import os
import asyncio
import aiohttp
from pytest import mark, fixture, raises
from proximl.exceptions import ApiError
from urllib.parse import urlparse

pytestmark = [mark.sdk, mark.integration, mark.jobs]


def extract_domain_suffix(hostname):
    parts = hostname.split(".")
    if len(parts) >= 2:
        return ".".join(parts[-2:])
    else:
        return None


@fixture(scope="class")
@mark.xdist_group("job_lifecycle")
async def job(proximl):
    job = await proximl.jobs.create(
        name="CLI Automated Tests - Job Lifecycle",
        type="notebook",
        gpu_types=["gtx1060"],
        gpu_count=1,
        disk_size=11,
        data=dict(datasets=[dict(id="CIFAR-10", public=True)]),
        model=dict(
            source_type="git",
            source_uri="git@github.com:proxiML/environment-tests.git",
        ),
    )
    yield job


@mark.create
@mark.asyncio
@mark.xdist_group("job_lifecycle")
class JobLifeCycleTests:
    async def test_wait_for_running(self, job):
        assert job.status != "running"
        job = await job.wait_for("running")
        assert job.status == "running"
        assert job.url
        assert extract_domain_suffix(urlparse(job.url).hostname) == "proximl.cloud"

    async def test_stop_job(self, job):
        assert job.status == "running"
        await job.stop()
        job = await job.wait_for("stopped", 120)
        assert job.status == "stopped"

    async def test_get_jobs(self, proximl, job):
        jobs = await proximl.jobs.list()
        assert len(jobs) > 0

    async def test_get_job(self, proximl, job):
        response = await proximl.jobs.get(job.id)
        assert response.id == job.id

    async def test_job_properties(self, job):
        assert isinstance(job.id, str)
        assert isinstance(job.name, str)
        assert isinstance(job.status, str)
        assert isinstance(job.type, str)

    async def test_job_str(self, job):
        string = str(job)
        regex = r"^{.*\"job_uuid\": \"" + job.id + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    async def test_job_repr(self, job):
        string = repr(job)
        regex = r"^Job\( proximl , {.*'job_uuid': '" + job.id + r"'.*}\)$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    async def test_start_job(self, job):
        assert job.status == "stopped"
        await job.start()
        job = await job.wait_for("running", 120)
        assert job.status == "running"

    async def test_copy_job_not_enough_disk(self, job):
        with raises(ApiError) as error:
            await job.copy(
                "CLI Automated Tests - Job Copy Not Enough Disk", disk_size=10
            )

        assert (
            "Invalid Request - Copied job disk_size must be greater than or equal to source job disk_size"
            in error.value.message
        )

    async def test_copy_job(self, job):
        await job.wait_for("running", 300)
        job_copy = await job.copy("CLI Automated Tests - Job Copy")
        assert job_copy.id != job.id
        await job_copy.wait_for("running", 300)
        assert job_copy.status == "running"
        await job_copy.stop()
        await job_copy.wait_for("stopped", 60)
        assert job_copy.status == "stopped"
        await job_copy.remove()

    async def test_convert_job(self, job):
        await job.wait_for("running", 300)
        training_job = await job.copy(
            name="CLI Automated Tests - Job Convert",
            type="training",
            workers=["python $ML_MODEL_PATH/tensorflow/main.py"],
            data=dict(
                datasets=[
                    dict(
                        id="CIFAR-10",
                        public=True,
                    )
                ],
                output_uri="s3://proximl-examples/output/resnet_cifar10",
                output_type="aws",
            ),
        )
        assert training_job.id
        training_job = await training_job.wait_for("finished", 300)
        assert training_job.status == "finished"
        await asyncio.sleep(10)  ## give billing a chance to update
        await training_job.refresh()
        assert training_job.credits > 0
        assert training_job.credits < 0.1
        await training_job.remove()

    async def test_remove_job(self, job):
        assert job.status == "running"
        await job.stop()
        await job.wait_for("stopped", 90)
        await job.remove()
        job = await job.wait_for("archived", 60)
        assert job is None


@mark.asyncio
class JobAPIResourceValidationTests:
    async def test_invalid_gpu_type(self, proximl):
        with raises(ApiError) as error:
            await proximl.jobs.create(
                name="Invalid GPU Type",
                type="training",
                gpu_types=["k80"],
                disk_size=10,
            )
        assert "Invalid Request - GPU Type k80 Invalid" in error.value.message

    async def test_invalid_disk_size(self, proximl):
        with raises(ApiError) as error:
            await proximl.jobs.create(
                name="Invalid Disk Size",
                type="training",
                gpu_types=["rtx3090"],
                disk_size=1,
            )
        assert (
            "Invalid Request - Disk Size must be between 10 and 2000"
            in error.value.message
        )

    async def test_combine_cpu_and_gpu_types(self, proximl):
        with raises(ApiError) as error:
            await proximl.jobs.create(
                name="Combine CPU and GPU Types",
                type="training",
                gpu_types=["cpu", "rtx3090"],
                disk_size=10,
            )
        assert (
            "Invalid Request - CPU Only may be not be combined with other GPU Types"
            in error.value.message
        )

    async def test_missing_cpu_count(self, proximl):
        with raises(ApiError) as error:
            await proximl.jobs.create(
                name="Missing CPU Count",
                type="training",
                gpu_types=["cpu"],
                disk_size=10,
            )
        assert (
            "Invalid Request - cpu_count required for CPU only jobs"
            in error.value.message
        )

    async def test_invalid_cpu_count(self, proximl):
        with raises(ApiError) as error:
            await proximl.jobs.create(
                name="Invalid CPU Count",
                type="training",
                gpu_types=["cpu"],
                cpu_count=1,
                disk_size=10,
            )
        assert (
            "Invalid Request - CPU Count must be a multiple of 4" in error.value.message
        )

    async def test_invalid_gpu_count_for_cpu(self, proximl):
        with raises(ApiError) as error:
            await proximl.jobs.create(
                name="Invalid GPU Count For CPU",
                type="training",
                gpu_types=["cpu"],
                gpu_count=1,
                cpu_count=4,
                disk_size=10,
            )
        assert (
            "Invalid Request - gpu_count not valid for CPU only jobs"
            in error.value.message
        )

    async def test_invalid_cpu_count_for_gpu(self, proximl):
        with raises(ApiError) as error:
            await proximl.jobs.create(
                name="Invalid CPU Count for GPU",
                type="notebook",
                gpu_types=["rtx3090"],
                cpu_count=4,
                disk_size=10,
            )
        assert (
            "Invalid Request - CPU Count must be at least 8 for gpu types rtx3090"
            in error.value.message
        )

    async def test_invalid_cpu_count_for_gpu_2(self, proximl):
        with raises(ApiError) as error:
            await proximl.jobs.create(
                name="Invalid CPU Count for GPU 2",
                type="notebook",
                gpu_types=["rtx2080ti"],
                gpu_count=2,
                cpu_count=4,
                disk_size=10,
            )
        assert (
            "Invalid Request - CPU Count must be at least 8 for gpu types rtx2080ti and gpu_count 2"
            in error.value.message
        )

    async def test_invalid_cpu_count_for_gpu_3(self, proximl):
        with raises(ApiError) as error:
            await proximl.jobs.create(
                name="Invalid CPU Count for GPU 3",
                type="training",
                gpu_types=["rtx2080ti", "rtx3090"],
                gpu_count=2,
                cpu_count=8,
                disk_size=10,
            )
        assert (
            "Invalid Request - CPU Count must be at least 16 for gpu types rtx2080ti,rtx3090 and gpu_count 2"
            in error.value.message
        )


@mark.asyncio
class JobAPIDataValidationTests:
    async def test_invalid_output_type_for_notebook(self, proximl):
        with raises(ApiError) as error:
            await proximl.jobs.create(
                name="Invalid Output Type for Notebook",
                type="notebook",
                gpu_types=["rtx3090"],
                disk_size=10,
                data=dict(
                    output_uri="s3://proximl-examples/output/resnet_cifar10",
                    output_type="aws",
                ),
            )
        assert (
            "Invalid Request - output_type invalid for Notebook and Endpoint jobs"
            in error.value.message
        )

    async def test_invalid_output_type_for_endpoint(self, proximl):
        with raises(ApiError) as error:
            await proximl.jobs.create(
                name="Invalid Output Type for Endpoint",
                type="endpoint",
                gpu_types=["rtx3090"],
                disk_size=10,
                endpoint=dict(
                    routes=[
                        dict(
                            path="/predict",
                            verb="POST",
                            function="predict_image",
                            file="predict",
                            positional=True,
                            body=[dict(name="filename", type="str")],
                        )
                    ]
                ),
                data=dict(
                    output_uri="s3://proximl-examples/output/resnet_cifar10",
                    output_type="aws",
                ),
            )
        assert (
            "Invalid Request - output_type invalid for Notebook and Endpoint jobs"
            in error.value.message
        )

    async def test_invalid_volumes_for_training(self, proximl):
        with raises(ApiError) as error:
            await proximl.jobs.create(
                name="Invalid Volumes for Training",
                type="training",
                gpu_types=["rtx3090"],
                disk_size=10,
                data=dict(
                    output_uri="s3://proximl-examples/output/resnet_cifar10",
                    output_type="aws",
                    volumes=["volume-id"],
                ),
                workers=["python train.py"],
            )
        assert (
            "Invalid Request - Only Notebook and Endpoint job types can use writable volumes"
            in error.value.message
        )

    async def test_invalid_volumes_for_inference(self, proximl):
        with raises(ApiError) as error:
            await proximl.jobs.create(
                name="Invalid Volumes for Inference",
                type="inference",
                gpu_types=["rtx3090"],
                disk_size=10,
                data=dict(
                    output_uri="s3://proximl-examples/output/resnet_cifar10",
                    output_type="aws",
                    volumes=["volume-id"],
                ),
                workers=["python predict.py"],
            )
        assert (
            "Invalid Request - Only Notebook and Endpoint job types can use writable volumes"
            in error.value.message
        )

    async def test_invalid_datasets_for_inference(self, proximl):
        with raises(ApiError) as error:
            await proximl.jobs.create(
                name="Datasets for Inference Job",
                type="inference",
                gpu_types=["rtx3090"],
                disk_size=10,
                data=dict(datasets=[dict(id="CIFAR-10", public=True)]),
                workers=["python predict.py"],
            )
        assert (
            "Invalid Request - Inference jobs cannot use datasets"
            in error.value.message
        )


@mark.asyncio
class JobAPIWorkerValidationTests:
    async def test_missing_workers_for_training(self, proximl):
        with raises(ApiError) as error:
            await proximl.jobs.create(
                name="Missing Workers for Training Job",
                type="training",
                gpu_types=["rtx3090"],
                disk_size=10,
            )
        assert (
            "Invalid Request - Training jobs must have at least one worker"
            in error.value.message
        )

    async def test_too_many_workers_for_inference(self, proximl):
        with raises(ApiError) as error:
            await proximl.jobs.create(
                name="Too Many Workers for Inference Job",
                type="inference",
                gpu_types=["rtx3090"],
                disk_size=10,
                workers=["python predict.py", "python predict.py"],
            )
        assert (
            "Invalid Request - Inference jobs must have exactly one worker"
            in error.value.message
        )

    async def test_invalid_worker_spec_for_endpoint(self, proximl):
        with raises(ApiError) as error:
            await proximl.jobs.create(
                name="Invalid Worker Spec for Endpoint",
                type="endpoint",
                gpu_types=["rtx3090"],
                disk_size=10,
                workers=["python predict.py"],
            )
        assert (
            "Invalid Request - Endpoints do not use worker commands"
            in error.value.message
        )


@mark.create
@mark.asyncio
@mark.xdist_group("job_io")
class JobIOTests:
    async def test_job_local_output(self, proximl, capsys):
        temp_dir = tempfile.TemporaryDirectory()
        job = await proximl.jobs.create(
            name="CLI Automated Tests - Local Output",
            type="training",
            gpu_types=["gtx1060"],
            disk_size=10,
            workers=["python $ML_MODEL_PATH/tensorflow/main.py"],
            environment=dict(
                type="DEEPLEARNING_PY312",
                env=[
                    dict(
                        key="CHECKPOINT_FILE",
                        value="model.ckpt-0050.weights.h5",
                    )
                ],
            ),
            data=dict(
                datasets=[
                    dict(
                        id="CIFAR-10",
                        public=True,
                    )
                ],
                output_uri=temp_dir.name,
                output_type="local",
            ),
            model=dict(
                source_type="git",
                source_uri="git@github.com:proxiML/environment-tests.git",
                checkpoints=[
                    "tensorflow-checkpoint",
                ],
            ),
        )
        await job.wait_for("waiting for data/model download")
        attach_task = asyncio.create_task(job.attach())
        connect_task = asyncio.create_task(job.connect())
        await asyncio.gather(attach_task, connect_task)
        await job.refresh()
        assert job.status == "finished"
        await job.disconnect()
        await job.remove()
        upload_contents = os.listdir(temp_dir.name)
        temp_dir.cleanup()
        assert any(
            "CLI_Automated_Tests_-_Local_Output" in content
            for content in upload_contents
        )

        captured = capsys.readouterr()
        sys.stdout.write(captured.out)
        sys.stderr.write(captured.err)
        assert "Epoch 1/2" in captured.out
        assert "Epoch 2/2" in captured.out
        assert "adding: model.ckpt-0001" in captured.out
        assert "Send complete" in captured.out

    async def test_job_model_input_and_output(self, proximl, capsys):
        model = await proximl.models.create(
            name="CLI Automated Tests - Job Git Model",
            source_type="git",
            source_uri="git@github.com:proxiML/environment-tests.git",
        )
        await model.wait_for("ready", 300)
        assert model.size >= 50000

        job = await proximl.jobs.create(
            "CLI Automated Tests - Training With proxiML Model Output",
            type="training",
            gpu_types=["gtx1060"],
            gpu_count=1,
            cpu_count=8,
            disk_size=10,
            worker_commands=["python $ML_MODEL_PATH/tensorflow/main.py"],
            data=dict(
                datasets=[
                    dict(
                        id="CIFAR-10",
                        public=True,
                    )
                ],
                output_type="proximl",
                output_uri="model",
            ),
            model=dict(source_type="proximl", source_uri=model.id),
        )
        await job.attach()
        await job.refresh()
        assert job.status == "finished"
        workers = job.workers
        await job.remove()
        await model.remove()
        captured = capsys.readouterr()
        sys.stdout.write(captured.out)
        sys.stderr.write(captured.err)
        assert "Epoch 1/2" in captured.out
        assert "Epoch 2/2" in captured.out

        new_model = await proximl.models.get(workers[0].get("output_uuid"))
        assert new_model.id
        await new_model.wait_for("ready")
        await new_model.refresh()
        assert new_model.size > model.size + 1000000
        assert (
            new_model.name
            == "Job - CLI Automated Tests - Training With proxiML Model Output"
        )
        await new_model.remove()


@mark.create
@mark.asyncio
@mark.xdist_group("job_types")
class JobTypeTests:
    async def test_endpoint(self, proximl):

        job = await proximl.jobs.create(
            "CLI Automated Tests - Endpoint",
            type="endpoint",
            gpu_type="GTX 1060",
            gpu_count=1,
            disk_size=10,
            model=dict(
                source_type="git",
                source_uri="https://github.com/proxiML/simple-tensorflow-classifier.git",
            ),
            endpoint=dict(
                routes=[
                    dict(
                        path="/predict",
                        verb="POST",
                        function="predict_image",
                        file="predict",
                        positional=True,
                        body=[dict(name="filename", type="str")],
                    )
                ]
            ),
        )
        await job.wait_for("running")
        await job.refresh()
        assert job.url
        assert extract_domain_suffix(urlparse(job.url).hostname) == "proximl.cloud"
        tries = 0
        await asyncio.sleep(180) ## downloading weights can be slow
        async with aiohttp.ClientSession() as session:
            retry = True
            while retry:
                async with session.request(
                    "GET",
                    f"{job.url}/ping",
                ) as resp:
                    if resp.status in [404, 502, 503]:
                        tries += 1
                        print(resp)
                        resp.close()
                        if tries == 5:
                            raise Exception("Too many errors")
                        else:
                            await asyncio.sleep(5)
                    else:
                        retry = False
                        results = await resp.json()
                        assert results["message"] == "pong"

        await job.stop()
        await job.wait_for("stopped")
        await job.refresh()
        assert not job.url
        await job.remove()

    async def test_job_custom_container(self, proximl, capsys):
        job = await proximl.jobs.create(
            name="CLI Automated Tests - Custom Container",
            type="training",
            gpu_types=["gtx1060", "rtx3090", "rtx2080ti"],
            gpu_count=1,
            disk_size=10,
            model=dict(
                source_type="git",
                source_uri="git@github.com:proxiML/environment-tests.git",
            ),
            environment=dict(
                type="CUSTOM",
                custom_image="tensorflow/tensorflow:2.13.0-gpu",
                packages=dict(
                    pip=[
                        "matplotlib",
                        "scipy",
                        "tensorflow_hub",
                        "keras_applications",
                        "keras_preprocessing",
                        "protobuf",
                        "typing-extensions",
                    ]
                ),
            ),
            worker_commands=[
                "python $ML_MODEL_PATH/tensorflow/main.py",
            ],
            data=dict(
                datasets=[
                    dict(
                        id="CIFAR-10",
                        public=True,
                    )
                ],
                output_type="wasabi",
                output_uri="s3://proximl-example/output/resnet_cifar10",
                output_options=dict(
                    endpoint_url="https://s3.wasabisys.com", archive=False
                ),
            ),
        )
        assert job.id
        await job.attach()
        await job.refresh()
        assert job.status == "finished"
        await job.remove()
        captured = capsys.readouterr()
        sys.stdout.write(captured.out)
        sys.stderr.write(captured.err)
        assert "Epoch 1/2" in captured.out
        assert "Epoch 2/2" in captured.out
        assert "Uploading s3://proximl-example/output/resnet_cifar10" in captured.out
        assert (
            "upload: ./model.ckpt-0002.weights.h5 to s3://proximl-example/output/resnet_cifar10/model.ckpt-0002.weights.h5"
            in captured.out
        )
        assert "Upload complete" in captured.out


@mark.create
@mark.asyncio
@mark.xdist_group("job_features")
class JobFeatureTests:
    async def test_cpu_instance(self, proximl, capsys):
        job = await proximl.jobs.create(
            name="CLI Automated Tests - CPU Instance",
            type="training",
            gpu_types=["cpu"],
            cpu_count=4,
            disk_size=10,
            model=dict(
                source_type="git",
                source_uri="git@github.com:proxiML/environment-tests.git",
            ),
            worker_commands=[
                "python $ML_MODEL_PATH/pytorch/main.py",
            ],
            data=dict(
                datasets=[dict(id="MNIST", public=True)],
            ),
        )
        assert job.id
        await job.attach()
        await job.refresh()
        assert job.status == "finished"
        await job.remove()
        captured = capsys.readouterr()
        sys.stdout.write(captured.out)
        sys.stderr.write(captured.err)
        assert "Train Epoch: 1 [0/60000 (0%)]" in captured.out
        assert "Train Epoch: 1 [59520/60000 (99%)]" in captured.out

    async def test_inference_job(self, proximl, capsys):
        temp_dir = tempfile.TemporaryDirectory()
        job = await proximl.jobs.create(
            name="CLI Automated Tests - Inference Job",
            type="inference",
            gpu_type="GTX 1060",
            gpu_count=1,
            disk_size=10,
            workers=[
                "python $ML_MODEL_PATH/tensorflow/main.py",
            ],
            data=dict(
                input_type="wasabi",
                input_uri="s3://proximl-example/input/cifar-10",
                input_options=dict(endpoint_url="https://s3.wasabisys.com"),
                output_type="local",
                output_uri=temp_dir.name,
                output_options=dict(archive=False),
            ),
            model=dict(git_uri="git@github.com:proxiML/environment-tests.git"),
        )
        assert job.id
        await job.wait_for("running")
        await job.connect()
        await job.attach()
        await job.refresh()
        assert job.status == "finished"
        await job.disconnect()
        await job.remove()
        await job.wait_for("archived")
        captured = capsys.readouterr()
        sys.stdout.write(captured.out)
        sys.stderr.write(captured.err)
        upload_contents = os.listdir(temp_dir.name)
        temp_dir.cleanup()
        assert len(upload_contents) >= 3
        assert any(
            "model.ckpt-0002" in content
            for content in upload_contents
        )

        captured = capsys.readouterr()
        sys.stdout.write(captured.out)
        sys.stderr.write(captured.err)
        assert "Epoch 1/2" in captured.out
        assert "Epoch 2/2" in captured.out
        assert "Number of regular files transferred: 4" in captured.out
        assert "Send complete" in captured.out



import re
import sys
import asyncio
from pytest import mark, fixture

pytestmark = [mark.sdk, mark.integration, mark.volumes]


@mark.create
@mark.asyncio
@mark.xdist_group("volumes")
class GetVolumeTests:
    @fixture(scope="class")
    async def volume(self, proximl):
        volume = await proximl.volumes.create(
            name="CLI Automated",
            source_type="git",
            source_uri="git@github.com:proxiML/environment-tests.git",
            capacity="10G",
        )
        volume = await volume.wait_for("ready", 120)
        yield volume
        await volume.remove()
        volume = await volume.wait_for("archived", 60)

    async def test_get_volumes(self, proximl, volume):
        volumes = await proximl.volumes.list()
        assert len(volumes) > 0

    async def test_get_volume(self, proximl, volume):
        response = await proximl.volumes.get(volume.id)
        assert response.id == volume.id

    async def test_volume_properties(self, volume):
        assert isinstance(volume.id, str)
        assert isinstance(volume.status, str)
        assert isinstance(volume.name, str)
        assert isinstance(volume.capacity, str)
        assert isinstance(volume.used_size, int)
        assert isinstance(volume.billed_size, int)

    async def test_volume_str(self, volume):
        string = str(volume)
        regex = r"^{.*\"id\": \"" + volume.id + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    async def test_volume_repr(self, volume):
        string = repr(volume)
        regex = r"^Volume\( proximl , \*\*{.*'id': '" + volume.id + r"'.*}\)$"
        assert isinstance(string, str)
        assert re.match(regex, string)


@mark.create
@mark.asyncio
async def test_volume_wasabi(proximl, capsys):
    volume = await proximl.volumes.create(
        name="CLI Automated Wasabi",
        source_type="wasabi",
        source_uri="s3://proximl-example/models/proximl-examples",
        capacity="10G",
        source_options=dict(endpoint_url="https://s3.wasabisys.com"),
    )
    volume = await volume.wait_for("ready", 300)
    status = volume.status
    billed_size = volume.billed_size
    used_size = volume.used_size
    await volume.remove()
    assert status == "ready"
    assert billed_size >= 10000000
    assert used_size >= 500000


@mark.create
@mark.asyncio
async def test_volume_local(proximl, capsys):
    volume = await proximl.volumes.create(
        name="CLI Automated Local",
        source_type="local",
        source_uri="~/tensorflow-model",
        capacity="10G",
    )
    attach_task = asyncio.create_task(volume.attach())
    connect_task = asyncio.create_task(volume.connect())
    await asyncio.gather(attach_task, connect_task)
    await volume.disconnect()
    await volume.refresh()
    status = volume.status
    billed_size = volume.billed_size
    used_size = volume.used_size
    await volume.remove()
    assert status == "ready"
    assert billed_size >= 10000000
    assert used_size >= 1000000
    captured = capsys.readouterr()
    sys.stdout.write(captured.out)
    sys.stderr.write(captured.err)
    assert "Starting data upload from local" in captured.out
    assert "official/LICENSE  11456 bytes" in captured.out
    assert "Upload complete" in captured.out



import re
import sys
import asyncio
from pytest import mark, fixture

pytestmark = [mark.sdk, mark.integration, mark.projects]


@mark.create
@mark.asyncio
@mark.xdist_group("project_resources")
class ProjectDatastoresTests:
    @fixture(scope="class")
    async def project_datastore(self, project):
        datastores = await project.datastores.list()
        yield datastores[0]

    async def test_list_project_datastores(self, project):
        datastores = await project.datastores.list()
        assert len(datastores) > 0

    async def test_project_datastore_properties(self, project, project_datastore):
        assert isinstance(project_datastore.project_uuid, str)
        assert isinstance(project_datastore.type, str)
        assert isinstance(project_datastore.name, str)
        assert isinstance(project_datastore.region_uuid, str)
        assert project.id == project_datastore.project_uuid

    async def test_project_datastore_str(self, project_datastore):
        string = str(project_datastore)
        regex = r"^{.*\"name\": \"" + project_datastore.name + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    async def test_project_datastore_repr(self, project_datastore):
        string = repr(project_datastore)
        regex = (
            r"^ProjectDatastore\( proximl , \*\*{.*'name': '"
            + project_datastore.name
            + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)



import re
import json
import os
from unittest.mock import (
    Mock,
    AsyncMock,
    patch,
    create_autospec,
    call,
    mock_open,
)
from pytest import mark, fixture
from aiodocker.containers import DockerContainer
import proximl.connections as specimen
from proximl.jobs import Job
from proximl.datasets import Dataset


pytestmark = [mark.sdk, mark.unit, mark.connections]


@fixture
def connections(mock_proximl):
    yield specimen.Connections(mock_proximl)


@fixture
def dataset_con(mock_proximl):
    DatasetMock = create_autospec(Dataset)
    yield specimen.Connection(
        mock_proximl,
        entity_type="dataset",
        id="data-id-1",
        entity=DatasetMock(
            mock_proximl,
            dataset_uuid="data-id-1",
            project_uuid="proj-id-1",
            name="first one",
            status="new",
        ),
    )


@fixture
def job_con(mock_proximl):
    JobMock = create_autospec(Job)
    yield specimen.Connection(
        mock_proximl,
        entity_type="job",
        id="job-id-1",
        entity=JobMock(
            mock_proximl,
            **{
                "customer_uuid": "cus-id-1",
                "project_uuid": "proj-id-1",
                "job_uuid": "job-id-1",
                "name": "test notebook",
                "type": "interactive",
                "status": "new",
            },
        ),
    )


class ConnectionsTests:
    @mark.asyncio
    async def test_connections_list(
        self,
        connections,
    ):
        with patch("proximl.connections.os") as mock_os:
            mock_os.listdir = Mock(
                return_value=["job_job-id-1", "dataset_data-id-1", "baddir"]
            )
            with patch(
                "proximl.connections.Connection", autospec=True
            ) as mock_connection:
                connection = mock_connection.return_value
                connections = await connections.list()
                connection.check.assert_called
                assert len(connections) == 2

    @mark.asyncio
    async def test_connections_cleanup(
        self,
        connections,
    ):
        with patch("proximl.connections.os") as mock_os:
            dir_list = ["job_job-id-1", "dataset_data-id-1"]
            mock_os.listdir = Mock(return_value=dir_list)

            with patch(
                "proximl.connections._cleanup_containers"
            ) as mock_cleanup:
                await connections.cleanup_connections()
                assert mock_cleanup.mock_calls == [
                    call(
                        "proj-id-1",
                        os.path.expanduser("~/.proximl/connections/proj-id-1"),
                        dir_list,
                        "vpn",
                    ),
                    call(
                        "proj-id-1",
                        os.path.expanduser("~/.proximl/connections/proj-id-1"),
                        dir_list,
                        "storage",
                    ),
                ]

    @mark.asyncio
    async def test_connections_cleanup_containers(
        self,
    ):
        with patch("proximl.connections.open", mock_open(read_data="keep-me")):

            with patch(
                "proximl.connections.aiodocker.Docker", autospec=True
            ) as mock_docker:
                containers = [
                    create_autospec(DockerContainer),
                    create_autospec(DockerContainer),
                ]
                containers[0].id = "keep-me"
                containers[1].id = "delete-me"
                docker = mock_docker.return_value
                docker.containers = AsyncMock()
                docker.containers.list = AsyncMock(return_value=containers)
                await specimen._cleanup_containers(
                    "proj-id-1",
                    os.path.expanduser("~/.proximl/connections/proj-id-1"),
                    ["dir_1"],
                    "job",
                )
                docker.containers.list.assert_called_once_with(
                    all=True,
                    filters=json.dumps(
                        dict(
                            label=[
                                "service=proximl",
                                "type=job",
                                "project=proj-id-1",
                            ]
                        )
                    ),
                )
                containers[0].delete.assert_not_called
                containers[1].delete.assert_called_once_with(force=True)


class ConnectionTests:
    def test_connection_properties(self, dataset_con):
        assert isinstance(dataset_con.id, str)
        assert isinstance(dataset_con.status, str)
        assert isinstance(dataset_con.type, str)
        assert dataset_con.type == "dataset"

    def test_connection_str(self, dataset_con):
        string = str(dataset_con)
        regex = (
            r"^Connection for "
            + dataset_con.type
            + " - "
            + dataset_con.id
            + ": "
            + dataset_con.status
            + r"$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_connection_repr(self, dataset_con):
        string = repr(dataset_con)
        regex = (
            r"^Connection\( proximl , "
            + dataset_con.id
            + ", "
            + dataset_con.type
            + r"\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)



import re
import sys
import asyncio
from pytest import mark, fixture

pytestmark = [mark.sdk, mark.integration, mark.checkpoints]


@mark.create
@mark.asyncio
@mark.xdist_group("checkpoints")
class GetCheckpointTests:
    @fixture(scope="class")
    async def checkpoint(self, proximl):
        checkpoint = await proximl.checkpoints.create(
            name="CLI Automated",
            source_type="git",
            source_uri="git@github.com:proxiML/environment-tests.git",
        )
        checkpoint = await checkpoint.wait_for("ready", 180)
        yield checkpoint
        await checkpoint.remove()
        checkpoint = await checkpoint.wait_for("archived", 60)

    async def test_get_checkpoints(self, proximl, checkpoint):
        checkpoints = await proximl.checkpoints.list()
        assert len(checkpoints) > 0

    async def test_get_checkpoint(self, proximl, checkpoint):
        response = await proximl.checkpoints.get(checkpoint.id)
        assert response.id == checkpoint.id

    async def test_checkpoint_properties(self, checkpoint):
        assert isinstance(checkpoint.id, str)
        assert isinstance(checkpoint.status, str)
        assert isinstance(checkpoint.name, str)
        assert isinstance(checkpoint.size, int)

    async def test_checkpoint_str(self, checkpoint):
        string = str(checkpoint)
        regex = r"^{.*\"checkpoint_uuid\": \"" + checkpoint.id + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    async def test_checkpoint_repr(self, checkpoint):
        string = repr(checkpoint)
        regex = (
            r"^Checkpoint\( proximl , \*\*{.*'checkpoint_uuid': '"
            + checkpoint.id
            + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)


@mark.create
@mark.asyncio
async def test_checkpoint_wasabi(proximl, capsys):
    checkpoint = await proximl.checkpoints.create(
        name="CLI Automated Wasabi",
        source_type="wasabi",
        source_uri="s3://proximl-example/models/proximl-examples",
        source_options=dict(endpoint_url="https://s3.wasabisys.com"),
    )
    checkpoint = await checkpoint.wait_for("ready", 300)
    status = checkpoint.status
    size = checkpoint.size
    await checkpoint.remove()
    assert status == "ready"
    assert size >= 500000


@mark.create
@mark.asyncio
async def test_checkpoint_local(proximl, capsys):
    checkpoint = await proximl.checkpoints.create(
        name="CLI Automated Local",
        source_type="local",
        source_uri="~/tensorflow-model",
    )
    attach_task = asyncio.create_task(checkpoint.attach())
    connect_task = asyncio.create_task(checkpoint.connect())
    await asyncio.gather(attach_task, connect_task)
    await checkpoint.disconnect()
    await checkpoint.refresh()
    status = checkpoint.status
    size = checkpoint.size
    await checkpoint.remove()
    assert status == "ready"
    assert size >= 1000000
    captured = capsys.readouterr()
    sys.stdout.write(captured.out)
    sys.stderr.write(captured.err)
    assert "Starting data upload from local" in captured.out
    assert "official/LICENSE  11456 bytes" in captured.out
    assert "Upload complete" in captured.out



import json
import logging


class ProjectDataConnectors(object):
    def __init__(self, proximl, project_id):
        self.proximl = proximl
        self.project_id = project_id

    async def get(self, id, **kwargs):
        resp = await self.proximl._query(
            f"/project/{self.project_id}/data_connectors/{id}", "GET", kwargs
        )
        return ProjectDataConnector(self.proximl, **resp)

    async def list(self, **kwargs):
        resp = await self.proximl._query(
            f"/project/{self.project_id}/data_connectors", "GET", kwargs
        )
        data_connectors = [
            ProjectDataConnector(self.proximl, **data_connector)
            for data_connector in resp
        ]
        return data_connectors

    async def refresh(self):
        await self.proximl._query(
            f"/project/{self.project_id}/data_connectors", "PATCH"
        )


class ProjectDataConnector:
    def __init__(self, proximl, **kwargs):
        self.proximl = proximl
        self._entity = kwargs
        self._id = self._entity.get("id")
        self._project_uuid = self._entity.get("project_uuid")
        self._name = self._entity.get("name")
        self._type = self._entity.get("type")
        self._region_uuid = self._entity.get("region_uuid")

    @property
    def id(self) -> str:
        return self._id

    @property
    def project_uuid(self) -> str:
        return self._project_uuid

    @property
    def name(self) -> str:
        return self._name

    @property
    def type(self) -> str:
        return self._type

    @property
    def region_uuid(self) -> str:
        return self._region_uuid

    def __str__(self):
        return json.dumps({k: v for k, v in self._entity.items()})

    def __repr__(self):
        return f"ProjectDataConnector( proximl , **{self._entity.__repr__()})"

    def __bool__(self):
        return bool(self._id)

    async def enable(self):
        await self.proximl._query(
            f"/project/{self._project_uuid}/data_connectors/{self._id}/enable", "PATCH"
        )

    async def disable(self):
        await self.proximl._query(
            f"/project/{self._project_uuid}/data_connectors/{self._id}/disable", "PATCH"
        )



import re
import logging
import json
from unittest.mock import AsyncMock, patch
from pytest import mark, fixture, raises
from aiohttp import WSMessage, WSMsgType

import proximl.jobs as specimen
from proximl.exceptions import (
    ApiError,
    JobError,
    SpecificationError,
    ProxiMLException,
)

pytestmark = [mark.sdk, mark.unit, mark.jobs]


@fixture
def jobs(mock_proximl):
    yield specimen.Jobs(mock_proximl)


@fixture
def job(mock_proximl):
    yield specimen.Job(
        mock_proximl,
        **{
            "customer_uuid": "cus-id-1",
            "project_uuid": "proj-id-1",
            "job_uuid": "job-id-1",
            "name": "test notebook",
            "start": "2021-02-11T15:46:22.455Z",
            "type": "notebook",
            "status": "new",
            "credits_per_hour": 0.1,
            "credits": 0.1007,
            "workers": [
                {
                    "rig_uuid": "rig-id-1",
                    "job_worker_uuid": "worker-id-1",
                    "command": "jupyter lab",
                    "status": "new",
                }
            ],
            "worker_status": "new",
            "resources": {
                "gpu_count": 1,
                "gpu_types": ["1060-id"],
                "gpu_type_id": "1060-id",
                "disk_size": 10,
                "max_price": 10,
            },
            "model": {
                "size": 7176192,
                "git_uri": "git@github.com:proxiML/test-private.git",
                "status": "new",
            },
            "data": {
                "datasets": [
                    {
                        "dataset_uuid": "data-id-1",
                        "name": "first one",
                        "type": "public",
                        "size": 184549376,
                    },
                    {
                        "dataset_uuid": "data-id-2",
                        "name": "second one",
                        "type": "public",
                        "size": 5068061409,
                    },
                ],
                "status": "ready",
            },
            "environment": {
                "type": "DEEPLEARNING_PY310",
                "image_size": 44966989795,
                "env": [
                    {"value": "env1val", "key": "env1"},
                    {"value": "env2val", "key": "env2"},
                ],
                "worker_key_types": ["aws", "gcp"],
                "status": "new",
            },
            "vpn": {
                "status": "new",
                "cidr": "10.106.171.0/24",
                "client": {
                    "port": "36017",
                    "id": "cus-id-1",
                    "address": "10.106.171.253",
                },
                "net_prefix_type_id": 1,
            },
            "nb_token": "token",
        },
    )


@fixture
def training_job(mock_proximl):
    yield specimen.Job(
        mock_proximl,
        **{
            "customer_uuid": "cus-id-1",
            "project_uuid": "proj-id-1",
            "job_uuid": "job-id-1",
            "name": "test training",
            "start": "2021-02-11T15:46:22.455Z",
            "type": "training",
            "status": "running",
            "credits_per_hour": 0.1,
            "credits": 0.1007,
            "workers": [
                {
                    "rig_uuid": "rig-id-1",
                    "job_worker_uuid": "worker-id-1",
                    "command": "python train.py",
                    "status": "running",
                }
            ],
            "worker_status": "running",
            "resources": {
                "gpu_count": 1,
                "gpu_type_id": "1060-id",
                "disk_size": 10,
                "max_price": 10,
            },
            "model": {
                "size": 7176192,
                "git_uri": "git@github.com:proxiML/test-private.git",
                "status": "new",
            },
            "data": {
                "datasets": [
                    {
                        "dataset_uuid": "data-id-1",
                        "name": "first one",
                        "type": "public",
                        "size": 184549376,
                    },
                    {
                        "dataset_uuid": "data-id-2",
                        "name": "second one",
                        "type": "public",
                        "size": 5068061409,
                    },
                ],
                "status": "ready",
            },
            "environment": {
                "type": "DEEPLEARNING_PY310",
                "image_size": 44966989795,
                "env": [
                    {"value": "env1val", "key": "env1"},
                    {"value": "env2val", "key": "env2"},
                ],
                "worker_key_types": ["aws", "gcp"],
                "status": "ready",
            },
            "vpn": {
                "status": "running",
                "cidr": "10.106.171.0/24",
                "client": {
                    "port": "36017",
                    "id": "cus-id-1",
                    "address": "10.106.171.253",
                },
            },
        },
    )


@mark.asyncio
class JobsTests:
    async def test_jobs_get(
        self,
        jobs,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await jobs.get("1234")
        mock_proximl._query.assert_called_once_with("/job/1234", "GET", dict())

    async def test_jobs_list(
        self,
        jobs,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await jobs.list()
        mock_proximl._query.assert_called_once_with("/job", "GET", dict())

    async def test_jobs_remove(
        self,
        jobs,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await jobs.remove("1234")
        mock_proximl._query.assert_called_once_with(
            "/job/1234", "DELETE", dict(force=True)
        )

    async def test_job_create_minimal(
        self,
        jobs,
        mock_proximl,
    ):
        requested_config = dict(
            name="job_name",
            type="notebook",
            gpu_type="GTX 1060",
            gpu_count=1,
            disk_size=10,
        )
        expected_payload = dict(
            project_uuid="proj-id-1",
            name="job_name",
            type="notebook",
            resources=dict(
                gpu_type_id="GTX 1060", gpu_count=1, disk_size=10, max_price=10
            ),
            model=dict(),
            worker_commands=[],
        )

        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await jobs.create(**requested_config)
        mock_proximl._query.assert_called_once_with(
            "/job", "POST", None, expected_payload
        )

    async def test_job_create_from_empty_copy(
        self,
        jobs,
        mock_proximl,
    ):
        requested_config = dict(
            name="job_name",
            type="notebook",
            gpu_type="1060-id",
            gpu_count=1,
            disk_size=10,
            max_price=2,
            worker_commands=None,
            environment=None,
            data=None,
            source_job_uuid="job-id-1",
        )
        expected_payload = dict(
            name="job_name",
            project_uuid="proj-id-1",
            type="notebook",
            resources=dict(
                gpu_type_id="1060-id", gpu_count=1, disk_size=10, max_price=2
            ),
            source_job_uuid="job-id-1",
        )

        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await jobs.create(**requested_config)
        mock_proximl._query.assert_called_once_with(
            "/job", "POST", None, expected_payload
        )

    async def test_job_missing_gpu_type(
        self,
        jobs,
    ):
        requested_config = dict(
            name="job_name",
            type="notebook",
            disk_size=10,
        )

        with raises(SpecificationError) as error:
            await jobs.create(**requested_config)
        assert (
            "Invalid resource specification, either 'gpu_type' or 'gpu_types' must be provided"
            in error.value.message
        )


class JobTests:
    def test_job_properties(self, job):
        assert isinstance(job.id, str)
        assert isinstance(job.name, str)
        assert isinstance(job.status, str)
        assert isinstance(job.type, str)

    def test_job_str(self, job):
        string = str(job)
        regex = r"^{.*\"job_uuid\": \"" + job.id + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_job_repr(self, job):
        string = repr(job)
        regex = r"^Job\( proximl , {.*'job_uuid': '" + job.id + r"'.*}\)$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_job_bool(self, job, mock_proximl):
        empty_job = specimen.Job(mock_proximl)
        assert bool(job)
        assert not bool(empty_job)

    @mark.asyncio
    async def test_job_start(self, job, mock_proximl):
        api_response = {
            "customer_uuid": "cus-id-1",
            "job_uuid": "job-id-1",
            "name": "test notebook",
            "type": "notebook",
            "status": "starting",
        }
        mock_proximl._query = AsyncMock(return_value=api_response)
        await job.start()
        mock_proximl._query.assert_called_once_with(
            "/job/job-id-1",
            "PATCH",
            dict(project_uuid="proj-id-1"),
            dict(command="start"),
        )

    @mark.asyncio
    async def test_job_stop(self, job, mock_proximl):
        api_response = {
            "customer_uuid": "cus-id-1",
            "job_uuid": "job-id-1",
            "name": "test notebook",
            "type": "notebook",
            "status": "stopping",
        }
        mock_proximl._query = AsyncMock(return_value=api_response)
        await job.stop()
        mock_proximl._query.assert_called_once_with(
            "/job/job-id-1",
            "PATCH",
            dict(project_uuid="proj-id-1"),
            dict(command="stop"),
        )

    @mark.asyncio
    async def test_job_get_worker_log_url(self, job, mock_proximl):
        api_response = "https://trainml-jobs-dev.s3.us-east-2.amazonaws.com/job-id-1/logs/worker-id-1/test_notebook.zip"
        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await job.get_worker_log_url("worker-id-1")
        mock_proximl._query.assert_called_once_with(
            "/job/job-id-1/worker/worker-id-1/logs",
            "GET",
            dict(project_uuid="proj-id-1"),
        )
        assert response == api_response

    @mark.asyncio
    async def test_job_get_connection_utility_url(self, job, mock_proximl):
        api_response = "https://trainml-jobs-dev.s3.us-east-2.amazonaws.com/job-id-1/vpn/proximl-test_notebook.zip"
        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await job.get_connection_utility_url()
        mock_proximl._query.assert_called_once_with(
            "/job/job-id-1/download", "GET", dict(project_uuid="proj-id-1")
        )
        assert response == api_response

    def test_job_get_connection_details_no_data(self, job):
        details = job.get_connection_details()
        expected_details = dict(
            project_uuid="proj-id-1",
            entity_type="job",
            cidr="10.106.171.0/24",
            ssh_port=None,
            model_path=None,
            input_path=None,
            output_path=None,
        )
        assert details == expected_details

    def test_job_get_connection_details_local_output_data(self, mock_proximl):
        job = specimen.Job(
            mock_proximl,
            **{
                "customer_uuid": "cus-id-1",
                "project_uuid": "proj-id-1",
                "job_uuid": "job-id-1",
                "name": "test notebook",
                "type": "notebook",
                "status": "new",
                "model": {},
                "data": {
                    "datasets": [],
                    "output_type": "local",
                    "output_uri": "~/tensorflow-example/output",
                    "status": "ready",
                },
                "vpn": {
                    "status": "new",
                    "cidr": "10.106.171.0/24",
                    "client": {
                        "port": "36017",
                        "id": "cus-id-1",
                        "address": "10.106.171.253",
                        "ssh_port": 46600,
                    },
                },
            },
        )
        details = job.get_connection_details()
        expected_details = dict(
            project_uuid="proj-id-1",
            entity_type="job",
            cidr="10.106.171.0/24",
            ssh_port=46600,
            model_path=None,
            input_path=None,
            output_path="~/tensorflow-example/output",
        )
        assert details == expected_details

    def test_job_get_connection_details_local_model_and_input_data(
        self, mock_proximl
    ):
        job = specimen.Job(
            mock_proximl,
            **{
                "customer_uuid": "cus-id-1",
                "project_uuid": "proj-id-1",
                "job_uuid": "job-id-1",
                "name": "test notebook",
                "type": "notebook",
                "status": "new",
                "model": {"source_type": "local", "source_uri": "~/model_dir"},
                "data": {
                    "datasets": [],
                    "input_type": "local",
                    "input_uri": "~/data_dir",
                    "status": "ready",
                },
                "vpn": {
                    "status": "new",
                    "cidr": "10.106.171.0/24",
                    "client": {
                        "port": "36017",
                        "id": "cus-id-1",
                        "address": "10.106.171.253",
                        "ssh_port": 46600,
                    },
                },
            },
        )
        details = job.get_connection_details()
        expected_details = dict(
            project_uuid="proj-id-1",
            entity_type="job",
            cidr="10.106.171.0/24",
            ssh_port=46600,
            model_path="~/model_dir",
            input_path="~/data_dir",
            output_path=None,
        )
        assert details == expected_details

    @mark.asyncio
    async def test_job_connect(self, training_job, mock_proximl):
        with patch(
            "proximl.jobs.Connection",
            autospec=True,
        ) as mock_connection:
            connection = mock_connection.return_value
            connection.status = "connected"
            resp = await training_job.connect()
            connection.start.assert_called_once()
            assert resp == "connected"

    @mark.asyncio
    async def test_job_disconnect(self, training_job, mock_proximl):
        with patch(
            "proximl.jobs.Connection",
            autospec=True,
        ) as mock_connection:
            connection = mock_connection.return_value
            connection.status = "removed"
            resp = await training_job.disconnect()
            connection.stop.assert_called_once()
            assert resp == "removed"

    @mark.asyncio
    async def test_job_remove(self, job, mock_proximl):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await job.remove()
        mock_proximl._query.assert_called_once_with(
            "/job/job-id-1",
            "DELETE",
            dict(project_uuid="proj-id-1", force=False),
        )

    @mark.asyncio
    async def test_job_refresh(self, job, mock_proximl):
        api_response = {
            "customer_uuid": "cus-id-1",
            "job_uuid": "job-id-1",
            "name": "test notebook",
            "type": "notebook",
            "status": "running",
        }
        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await job.refresh()
        mock_proximl._query.assert_called_once_with(
            "/job/job-id-1", "GET", dict(project_uuid="proj-id-1")
        )
        assert job.status == "running"
        assert response.status == "running"

    def test_job_default_ws_msg_handler(self, job, capsys):
        data = {
            "msg": "Epoch (1/1000)\n",
            "time": 1613079345318,
            "type": "subscription",
            "stream": "worker-id-1",
            "job_worker_uuid": "worker-id-1",
        }

        handler = job._get_msg_handler(None)
        handler(data)
        captured = capsys.readouterr()
        assert captured.out == "02/11/2021, 15:35:45: Epoch (1/1000)\n"

    def test_job_default_ws_msg_handler_multiple_workers(
        self, mock_proximl, capsys
    ):
        job = specimen.Job(
            mock_proximl,
            **{
                "customer_uuid": "cus-id-1",
                "job_uuid": "job-id-1",
                "name": "test notebook",
                "type": "notebook",
                "status": "running",
                "workers": [
                    {
                        "rig_uuid": "rig-id-1",
                        "job_worker_uuid": "worker-id-1",
                        "command": "jupyter lab",
                        "status": "running",
                    },
                    {
                        "rig_uuid": "rig-id-1",
                        "job_worker_uuid": "worker-id-2",
                        "command": "jupyter lab",
                        "status": "running",
                    },
                ],
            },
        )

        data = {
            "msg": "Epoch (1/1000)\n",
            "time": 1613079345318,
            "type": "subscription",
            "stream": "worker-id-1",
            "job_worker_uuid": "worker-id-1",
        }

        handler = job._get_msg_handler(None)
        handler(data)
        captured = capsys.readouterr()
        assert (
            captured.out == "02/11/2021, 15:35:45: Worker 1 - Epoch (1/1000)\n"
        )

    def test_job_custom_ws_msg_handler(self, job, capsys):
        def custom_handler(msg):
            print(msg.get("stream"))

        data = {
            "msg": "Epoch (1/1000)\n",
            "time": 1613079345318,
            "type": "subscription",
            "stream": "worker-id-1",
            "job_worker_uuid": "worker-id-1",
        }

        handler = job._get_msg_handler(custom_handler)
        handler(data)
        captured = capsys.readouterr()
        assert captured.out == "worker-id-1\n"

    @mark.asyncio
    async def test_notebook_attach_error(self, job, mock_proximl):
        api_response = None
        mock_proximl._ws_subscribe = AsyncMock(return_value=api_response)
        with raises(SpecificationError):
            await job.attach()
        mock_proximl._ws_subscribe.create.assert_not_called()

    @mark.asyncio
    async def test_job_attach(self, mock_proximl):
        job_spec = {
            "customer_uuid": "cus-id-1",
            "job_uuid": "job-id-1",
            "name": "test training",
            "type": "training",
            "status": "running",
            "workers": [
                {
                    "rig_uuid": "rig-id-1",
                    "job_worker_uuid": "worker-id-1",
                    "command": "jupyter lab",
                    "status": "running",
                },
                {
                    "rig_uuid": "rig-id-1",
                    "job_worker_uuid": "worker-id-2",
                    "command": "jupyter lab",
                    "status": "running",
                },
            ],
        }
        job = specimen.Job(
            mock_proximl,
            **job_spec,
        )
        api_response = None
        mock_proximl._ws_subscribe = AsyncMock(return_value=api_response)
        job.refresh = AsyncMock(return_value=job_spec)
        await job.attach()
        mock_proximl._ws_subscribe.assert_called_once()

    @mark.asyncio
    async def test_job_attach_immediate_return(self, mock_proximl):
        job_spec = {
            "customer_uuid": "cus-id-1",
            "job_uuid": "job-id-1",
            "name": "test training",
            "type": "training",
            "status": "finished",
            "workers": [],
        }
        job = specimen.Job(
            mock_proximl,
            **job_spec,
        )
        api_response = None
        mock_proximl._ws_subscribe = AsyncMock(return_value=api_response)
        job.refresh = AsyncMock(return_value=job_spec)
        await job.attach()
        mock_proximl._ws_subscribe.assert_not_called()

    @mark.asyncio
    async def test_job_copy_minimal(
        self,
        job,
        mock_proximl,
    ):
        requested_config = dict(
            name="job_copy",
        )
        expected_kwargs = {
            "type": "notebook",
            "gpu_type": None,
            "gpu_types": ["1060-id"],
            "gpu_count": 1,
            "cpu_count": None,
            "disk_size": 10,
            "max_price": 10,
            "worker_commands": None,
            "workers": None,
            "environment": None,
            "data": None,
            "model": None,
            "source_job_uuid": "job-id-1",
        }

        api_response = specimen.Job(
            mock_proximl,
            **{
                "customer_uuid": "cus-id-1",
                "job_uuid": "job-id-2",
                "name": "job_copy",
                "type": "notebook",
                "status": "new",
            },
        )
        mock_proximl.jobs.create = AsyncMock(return_value=api_response)
        new_job = await job.copy(**requested_config)
        mock_proximl.jobs.create.assert_called_once_with(
            "job_copy", **expected_kwargs
        )
        assert new_job.id == "job-id-2"

    @mark.asyncio
    async def test_job_copy_training_job_failure(
        self,
        mock_proximl,
    ):
        job = specimen.Job(
            mock_proximl,
            **{
                "customer_uuid": "cus-id-1",
                "job_uuid": "job-id-1",
                "name": "test training",
                "type": "training",
                "status": "running",
            },
        )
        mock_proximl.jobs.create = AsyncMock()
        with raises(SpecificationError):
            await job.copy("new job")
        mock_proximl.jobs.create.assert_not_called()

    @mark.asyncio
    async def test_job_wait_for_successful(self, job, mock_proximl):
        api_response = {
            "customer_uuid": "cus-id-1",
            "job_uuid": "job-id-1",
            "name": "test notebook",
            "type": "notebook",
            "status": "running",
        }
        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await job.wait_for("running")
        mock_proximl._query.assert_called_once_with(
            "/job/job-id-1", "GET", dict(project_uuid="proj-id-1")
        )
        assert job.status == "running"
        assert response.status == "running"

    @mark.asyncio
    async def test_job_wait_for_current_status(self, mock_proximl):
        job = specimen.Job(
            mock_proximl,
            **{
                "customer_uuid": "cus-id-1",
                "job_uuid": "job-id-1",
                "name": "test notebook",
                "type": "notebook",
                "status": "running",
            },
        )
        api_response = None
        mock_proximl._query = AsyncMock(return_value=api_response)
        await job.wait_for("running")
        mock_proximl._query.assert_not_called()

    @mark.asyncio
    async def test_job_wait_for_incorrect_status(self, job, mock_proximl):
        api_response = None
        mock_proximl._query = AsyncMock(return_value=api_response)
        with raises(SpecificationError):
            await job.wait_for("ready")
        mock_proximl._query.assert_not_called()

    @mark.asyncio
    async def test_job_wait_for_with_delay(self, job, mock_proximl):
        api_response_initial = {
            "customer_uuid": "cus-id-1",
            "job_uuid": "job-id-1",
            "name": "test notebook",
            "type": "notebook",
            "status": "new",
        }
        api_response_final = {
            "customer_uuid": "cus-id-1",
            "job_uuid": "job-id-1",
            "name": "test notebook",
            "type": "notebook",
            "status": "running",
        }
        mock_proximl._query = AsyncMock()
        mock_proximl._query.side_effect = [
            api_response_initial,
            api_response_initial,
            api_response_final,
        ]
        response = await job.wait_for("running")
        assert job.status == "running"
        assert response.status == "running"

    @mark.asyncio
    async def test_job_wait_for_timeout(self, job, mock_proximl):
        api_response = {
            "customer_uuid": "cus-id-1",
            "job_uuid": "job-id-1",
            "name": "test notebook",
            "type": "notebook",
            "status": "new",
        }
        mock_proximl._query = AsyncMock(return_value=api_response)
        with raises(ProxiMLException):
            await job.wait_for("running", 10)
        mock_proximl._query.assert_called()

    @mark.asyncio
    async def test_job_wait_for_job_failed(self, job, mock_proximl):
        api_response = {
            "customer_uuid": "cus-id-1",
            "job_uuid": "job-id-1",
            "name": "test notebook",
            "type": "notebook",
            "status": "failed",
        }
        mock_proximl._query = AsyncMock(return_value=api_response)
        with raises(JobError):
            await job.wait_for("running")
        mock_proximl._query.assert_called()

    @mark.asyncio
    async def test_job_wait_for_archived_succeeded(self, job, mock_proximl):
        mock_proximl._query = AsyncMock(
            side_effect=ApiError(404, dict(errorMessage="Job Not Found"))
        )
        await job.wait_for("archived")
        mock_proximl._query.assert_called()

    @mark.asyncio
    async def test_job_wait_for_unexpected_api_error(self, job, mock_proximl):
        mock_proximl._query = AsyncMock(
            side_effect=ApiError(404, dict(errorMessage="Job Not Found"))
        )
        with raises(ApiError):
            await job.wait_for("running")
        mock_proximl._query.assert_called()



import re
import json
import click
from unittest.mock import AsyncMock, patch
from pytest import mark, fixture, raises

pytestmark = [mark.cli, mark.unit, mark.cloudbender, mark.services]

from proximl.cli.cloudbender import service as specimen
from proximl.cloudbender.services import Service


def test_list(runner, mock_services):
    with patch("proximl.cli.ProxiML", new=AsyncMock) as mock_proximl:
        mock_proximl.cloudbender = AsyncMock()
        mock_proximl.cloudbender.services = AsyncMock()
        mock_proximl.cloudbender.services.list = AsyncMock(return_value=mock_services)
        result = runner.invoke(
            specimen,
            args=["list", "--provider=prov-id-1", "--region=reg-id-1"],
        )
        assert result.exit_code == 0
        mock_proximl.cloudbender.services.list.assert_called_once_with(
            provider_uuid="prov-id-1", region_uuid="reg-id-1"
        )


def test_list_no_provider(runner, mock_services):
    with patch("proximl.cli.ProxiML", new=AsyncMock) as mock_proximl:
        mock_proximl.cloudbender = AsyncMock()
        mock_proximl.cloudbender.services = AsyncMock()
        mock_proximl.cloudbender.services.list = AsyncMock(return_value=mock_services)
        result = runner.invoke(specimen, ["list"])
        assert result.exit_code != 0



import re
import sys
import asyncio
from pytest import mark, fixture

pytestmark = [mark.sdk, mark.integration, mark.projects]


@mark.create
@mark.asyncio
@mark.xdist_group("project_resources")
class ProjectSecretsTests:
    @fixture(scope="class")
    async def project_secret(self, project):
        project_secret = await project.secrets.put(
            name="secret_value", value="IUHKLHKAHF"
        )
        yield project_secret
        await project.secrets.remove(name="secret_value")

    async def test_list_project_secrets(self, project, project_secret):
        secrets = await project.secrets.list()
        assert len(secrets) > 0

    async def test_project_secret_properties(self, project, project_secret):
        assert isinstance(project_secret.project_uuid, str)
        assert isinstance(project_secret.name, str)
        assert project_secret.name == "secret_value"
        assert project.id == project_secret.project_uuid

    async def test_project_secret_str(self, project_secret):
        string = str(project_secret)
        regex = r"^{.*\"name\": \"" + project_secret.name + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    async def test_project_secret_repr(self, project_secret):
        string = repr(project_secret)
        regex = (
            r"^ProjectSecret\( proximl , \*\*{.*'name': '"
            + project_secret.name
            + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)
